#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

if __name__ == '__main__':  # pragma: no cover
    import sys
    import os
    import subprocess
    import logging
    import platform

    libpaths = [
        "/Library/cloudprint-cups/",
        "/usr/local/share/cloudprint-cups/",
        "/usr/share/cloudprint-cups"
    ]
    addedPath = False
    for libpath in libpaths:
        if os.path.exists(libpath):
            sys.path.insert(0, libpath)
            addedPath = True
            break

    if not addedPath:
        sys.stderr.write("ERROR: Could not find any valid path for python files\n")
        sys.exit(1)

    from auth import Auth
    from printermanager import PrinterManager
    from ccputils import Utils

    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    if len(sys.argv) != 1 and len(sys.argv) < 6 or len(sys.argv) > 7:
        sys.stderr.write(
            "ERROR: Usage: %s job-id user title copies options [file]\n" % sys.argv[0])
        sys.exit(0)

    if len(sys.argv) >= 4 and sys.argv[3] == "Set Default Options":
        sys.stderr.write("ERROR: Unimplemented command: " + sys.argv[3] + "\n")
        logging.error("Unimplemented command: " + sys.argv[3])
        sys.exit(0)

    printFile = None

    if len(sys.argv) == 7:
        sys.stderr.write("ERROR: Sorry, CUPS Cloud Print no longer supports printing\
                          files directly for security reasons\n")
        sys.exit(1)
    if len(sys.argv) == 6:
        prog, jobID, userName, jobTitle, copies, printOptions = sys.argv
        printFile = jobTitle

    requestors, storage = Auth.SetupAuth(False)
    if not requestors:
        sys.stderr.write("ERROR: config is invalid or missing\n")
        logging.error("backend tried to run with invalid config")
        sys.exit(1)
    printer_manager = PrinterManager(requestors)

    if len(sys.argv) == 1:
        print 'network ' + Utils._PROTOCOL_NAME + ' "Unknown" "Google Cloud Print"'

        printers = printer_manager.getPrinters()
        if printers is not None:
            try:
                for printer in printers:
                    print printer.getCUPSBackendDescription()
            except Exception as error:
                sys.stderr.write("ERROR: " + error)
                logging.error(error)
                sys.exit(1)
        sys.exit(0)

    filedata = ""
    for line in sys.stdin:
        filedata += line

    # Backends should only produce multiple copies if a file name is
    # supplied (see CUPS Software Programmers Manual)
    copies = 1

    uri = os.getenv('DEVICE_URI')
    cupsprintername = os.getenv('PRINTER')
    if uri is None:
        message = 'URI must be "' + Utils._PROTOCOL + '<account name>/<cloud printer id>"!\n'
        sys.stdout.write(message)
        sys.exit(255)

    logging.info("Printing file " + str(printFile))
    optionsstring = ' '.join(["'%s'" % option for option in sys.argv])
    logging.info("Device is %s , printername is %s, params are: %s" %
                 (uri, cupsprintername, optionsstring))

    # setup
    useTempFile = False
    convertToPDFParams = ["ps2pdf", "-dPDFSETTINGS=/printer",
                          "-dUseCIEColor", "-", "-"]
    if Utils.which("ps2pdf") is None:
        convertToPDFParams = [Utils.which("pstopdf"), "-", "-o", "-"]
        if platform.system() == 'Darwin':
            useTempFile = True
    else:
        convertToPDFParams[0] = Utils.which("ps2pdf")

    logging.debug('is this a pdf? ' + str(printFile))
    result = 0

    if not Utils.fileIsPDF(filedata):
        tempFileNameIn = None
        tempFileNameOut = None
        if useTempFile:
            tempFileNameIn = Utils.GetTempFileName()
            if not Utils.WriteFile(tempFileNameIn, filedata):
                sys.stderr.write("ERROR: Failed to write file %s \n" % tempFileNameIn)
                sys.exit(1)
            convertToPDFParams[1] = tempFileNameIn

            tempFileNameOut = Utils.GetTempFileName()
            convertToPDFParams[2] = '-o'
            convertToPDFParams[3] = tempFileNameOut
            sys.stderr.write("INFO: Using temp files: %s %s\n" % (tempFileNameIn, tempFileNameOut))
            sys.stderr.write("INFO: Command is: %s\n" % (" ".join(convertToPDFParams)))

        # read file as pdf
        sys.stderr.write("INFO: Converting print job to PDF\n")
        p = subprocess.Popen(convertToPDFParams, stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                             env={'PATH': Utils.getPath()})
        processdata = p.communicate(filedata)
        if useTempFile:
            filedata = Utils.ReadFile(tempFileNameOut)
            if filedata is None:
                sys.stderr.write("ERROR: Failed to read file %s\n" % tempFileNameOut)
                sys.exit(1)
        else:
            filedata = processdata[0]
        if p.returncode != 0:
            sys.stderr.write("ERROR: Failed to convert file to pdf, returncode: %s\n" %
                             str(p.returncode))
            logging.error("Using these params %s", " ".join(convertToPDFParams))
            logging.error("Error from converting file to pdf: %s" % processdata[1])
            result = 1
        else:
            logging.info("Converted to PDF - %s bytes" % str(len(filedata)))

        if useTempFile:
            os.unlink(tempFileNameIn)
            os.unlink(tempFileNameOut)
    else:
        # read file normally
        logging.info("Using %s as is already PDF - %s bytes" % (printFile, len(filedata)))

    # send pdf data to GCP
    if result == 0:
        sys.stderr.write("INFO: Sending document to Cloud Print\n")
        logging.info("Sending %s to cloud" % printFile)

        printer = printer_manager.getPrinterByURI(uri)
        if printer is None:
            sys.stderr.write("ERROR: PrinterManager '%s' not found\n" % uri)
            result = 1
        elif printer.submitJob('pdf', printFile, filedata, jobTitle, cupsprintername, printOptions):
            sys.stderr.write("INFO: Successfully printed\n")
            result = 0
        else:
            sys.stderr.write("ERROR: Failed to submit job to cloud print\n")
            result = 1

        logging.info(str(printFile) + " sent to cloud print")

        if result != 0:
            sys.stderr.write("INFO: Printing Failed\n")
            logging.info("Failed printing")
        else:
            sys.stderr.write("INFO: Printing Successful\n")
            logging.info("Completed printing")

    sys.exit(result)



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import sys
sys.path.insert(0, ".")
from cloudprintrequestor import CloudPrintRequestor

class MockRequestor(CloudPrintRequestor):

    printers = []

    def mockSearch(self, path, headers, data, boundary):
        return json.dumps({'printers': self.printers})

    def mockSubmit(self, path, headers, data, boundary):
        if 'FAIL PAGE' in data:
            result = {
                'success': False,
                'message': 'FAIL PAGE was in message'}
        elif 'TEST PAGE WITH EXCEPTION' in data:
            raise Exception("Test exception")
        else:   
            result = {'success': True}
        return json.dumps(result)

    def mockPrinter(self, path, headers, data, boundary):
        printername = path.split('=')[1].split('&')[0]
        foundPrinter = None
        for printer in self.printers:
            if printer['id'] == printername:
                foundPrinter = printer
                break

        if foundPrinter is None:
            return json.dumps(None)

        result = {'success' : True, 'printers': [foundPrinter]}
        return json.dumps(result)

    def doRequest(self, path, headers=None, data=None, boundary=None):
        if (path.startswith('search?')):
            return json.loads(self.mockSearch(path, headers, data, boundary))
        if (path.startswith('printer?')):
            return json.loads(self.mockPrinter(path, headers, data, boundary))
        if (path.startswith('submit')):
            return json.loads(self.mockSubmit(path, headers, data, boundary))
        return None



#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

if __name__ == '__main__':  # pragma: no cover
    from auth import Auth
    from printermanager import PrinterManager

    from ccputils import Utils
    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    while True:
        result, storage = Auth.SetupAuth(False)
        if not result:
            print "No accounts are currently setup"
            break
        else:
            requestors = result
            print "You currently have these accounts configured: "
            i = 0
            accounts = []
            for requestor in requestors:
                i += 1
                accounts.append(requestor.getAccount())
                print str(i) + ") " + requestor.getAccount()
            print "0) Exit"
            answer = raw_input("Which account to delete (1-" + str(i) + ") ? ")
            if (answer.isdigit() and int(answer) <= i and int(answer) >= 1):
                if (Auth.DeleteAccount(accounts[int(answer) - 1]) is None):
                    print accounts[int(answer) - 1] + " deleted."
                    deleteprintersanswer = raw_input(
                        "Also delete associated printers? ")
                    if deleteprintersanswer.lower().startswith("y"):
                        printer_manager = PrinterManager(requestors)
                        printers, connection = \
                            printer_manager.getCUPSPrintersForAccount(accounts[int(answer) - 1])
                        if len(printers) == 0:
                            print "No printers to delete"
                        else:
                            for cupsPrinter in printers:
                                print "Deleting " + cupsPrinter['printer-info']
                                deleteReturnValue = connection.deletePrinter(
                                    cupsPrinter['printer-info'])
                                if deleteReturnValue is not None:
                                    errormessage = "Error deleting printer: "
                                    errormessage += str(deleteReturnValue)
                                    print errormessage
                    else:
                        print "Not deleting associated printers"
                else:
                    errormessage = "Error deleting stored "
                    errormessage += "credentials, perhaps "
                    errormessage += Auth.config + " is not writable?"
                    print errormessage
            elif (answer == "0"):
                break
            else:
                print "Invalid response, use '0' to exit"



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import sys
# workaround for ubuntu 12.04 / older python-six version
try:
    from six.moves import urllib
except ImportError:
    import six
    import urllib
    import urlparse
    six.moves.urllib = urllib
    six.moves.urllib.parse = urlparse
    six.moves.urllib.parse.urlencode = urllib.urlencode

from oauth2client import client
from oauth2client import multistore_file
from cloudprintrequestor import CloudPrintRequestor
from ccputils import Utils
from oauth2client.client import AccessTokenRefreshError


class Auth:
    clientid = "843805314553.apps.googleusercontent.com"
    clientsecret = 'MzTBsY4xlrD_lxkmwFbBrvBv'
    config = '/etc/cloudprint.conf'

    @staticmethod
    def RenewToken(interactive, requestor, credentials, storage, userid):
        try:
            credentials.refresh(requestor)
        except AccessTokenRefreshError as e:
            if not interactive:
                message = "ERROR: Failed to renew token "
                message += "(error: "
                message += str(e)
                message += "), "
                message += "please re-run "
                message += "/usr/share/cloudprint-cups/"
                message += "setupcloudprint.py\n"
                sys.stderr.write(message)
                sys.exit(1)
            else:
                message = "Failed to renew token (error: "
                message += str(e) + "), "
                message += "authentication needs to be "
                message += "setup again:\n"
                sys.stderr.write(message)
                Auth.AddAccount(storage, userid)
                credentials = storage.get()
        return credentials

    @staticmethod
    def DeleteAccount(userid=None):
        """Delete an account from the configuration file

        Args:
          storage: storage, instance of storage to store credentials in.
          userid: string, reference for the account

        Returns:
          deleted: boolean , true on success
        """
        storage = multistore_file.get_credential_storage(
            Auth.config,
            Auth.clientid,
            userid,
            ['https://www.googleapis.com/auth/cloudprint'])
        return storage.delete()

    @staticmethod
    def AddAccount(storage, userid=None,
                   permissions=None):
        """Adds an account to the configuration file

        Args:
          storage: storage, instance of storage to store credentials in.
          userid: string, reference for the account

        Returns:
          credentials: A credentials instance with the account details
        """
        if permissions is None:
            permissions = ['https://www.googleapis.com/auth/cloudprint']

        if userid is None:
            userid = raw_input(
                "Name for this user account ( eg something@gmail.com )? ")

            # setup storage again if just got userid now
            storage = multistore_file.get_credential_storage(
                Auth.config,
                Auth.clientid,
                userid,
                permissions)

        while True:
            flow = client.OAuth2WebServerFlow(client_id=Auth.clientid,
                                              client_secret=Auth.clientsecret,
                                              scope=permissions,
                                              user_agent=userid,
                                              redirect_uri='urn:ietf:wg:oauth:2.0:oob')
            auth_uri = flow.step1_get_authorize_url()
            message = "Open this URL, grant access to CUPS Cloud Print,"
            message += "then provide the code displayed : \n\n"
            message += auth_uri + "\n"
            print message
            code = raw_input('Code from Google: ')
            try:
                print ""
                credentials = flow.step2_exchange(code)
                storage.put(credentials)

                # fix permissions
                Utils.FixFilePermissions(Auth.config)

                return credentials
            except Exception as e:
                message = "\nThe code does not seem to be valid ( "
                message += str(e) + " ), please try again.\n"
                print message

    @staticmethod
    def SetupAuth(interactive=False,
                  permissions=None):
        """Sets up requestors with authentication tokens

        Args:
          interactive: boolean, when set to true can prompt user, otherwise
                       returns False if authentication fails

        Returns:
          requestor, storage: Authenticated requestors and an instance
                              of storage
        """
        if permissions is None:
            permissions = ['https://www.googleapis.com/auth/cloudprint']
        modifiedconfig = False

        # parse config file and extract useragents, which we use for account
        # names
        userids = []
        if os.path.exists(Auth.config):
            content_file = open(Auth.config, 'r')
            content = content_file.read()
            data = json.loads(content)
            for user in data['data']:
                userids.append(str(user['credential']['user_agent']))
        else:
            modifiedconfig = True

        if len(userids) == 0:
            userids = [None]

        requestors = []
        for userid in userids:
            storage = multistore_file.get_credential_storage(
                Auth.config,
                Auth.clientid,
                userid,
                permissions)
            credentials = storage.get()

            if not credentials and interactive:
                credentials = Auth.AddAccount(storage, userid, permissions)
                modifiedconfig = True
                if userid is None:
                    userid = credentials.user_agent

            if credentials:
                # renew if expired
                requestor = CloudPrintRequestor()
                if credentials.access_token_expired:
                    Auth.RenewToken(interactive, requestor, credentials, storage, userid)
                requestor = credentials.authorize(requestor)
                requestor.setAccount(userid)
                requestors.append(requestor)

        # fix permissions
        if modifiedconfig:
            Utils.FixFilePermissions(Auth.config)

        if not credentials:
            return False, False
        else:
            return requestors, storage

    @staticmethod
    def GetAccountNames(requestors):
        requestorAccounts = []
        for requestor in requestors:
            requestorAccounts.append(requestor.getAccount())
        return requestorAccounts



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2014 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import os
import logging
import sys
import grp
import base64
import fcntl
import termios
import struct
import string
import random
import tempfile


class Utils:
    logpath = '/var/log/cups/cloudprint_log'

    # Countries where letter sized paper is used, according to:
    # http://en.wikipedia.org/wiki/Letter_(paper_size)
    _LETTER_COUNTRIES = set(('US', 'CA', 'MX', 'BO', 'CO', 'VE', 'PH', 'CL'))
    _PROTOCOL_NAME = 'gcp'
    _PROTOCOL = _PROTOCOL_NAME + '://'
    _OLD_PROTOCOL_NAME = 'cloudprint'
    _OLD_PROTOCOL = _OLD_PROTOCOL_NAME + '://'
    _MIMETYPES_JOBTYPES = {'pdf': 'application/pdf',
                           'other': 'application/octet-stream',
                           'jpg': 'image/jpeg',
                           'png': 'image/png'}

    @staticmethod
    def FixFilePermissions(filename):
        filePermissions = True
        fileOwnerships = True
        currentStat = None
        if os.path.exists(filename):
            currentStat = os.stat(filename)

        if currentStat is None or currentStat.st_mode != 0o100660:
            try:
                os.chmod(filename, 0o100660)
            except Exception:
                filePermissions = False
                sys.stderr.write(
                    "DEBUG: Cannot alter " +
                    filename +
                    " file permissions\n")
                pass

        if currentStat is None or currentStat.st_gid != Utils.GetLPID():
            try:
                os.chown(filename, -1, Utils.GetLPID())
            except Exception:
                fileOwnerships = False
                sys.stderr.write(
                    "DEBUG: Cannot alter " +
                    filename +
                    " file ownership\n")
                pass

        return filePermissions, fileOwnerships

    @staticmethod
    def SetupLogging(logpath=None):
        returnValue = True
        logformat = "%(asctime)s|%(levelname)s|%(message)s"
        dateformat = "%Y-%m-%d %H:%M:%S"
        if logpath is None:
            logpath = Utils.logpath
        try:
            logging.basicConfig(
                filename=logpath,
                level=logging.INFO,
                format=logformat,
                datefmt=dateformat)
            Utils.FixFilePermissions(logpath)
        except Exception:
            logging.basicConfig(
                level=logging.INFO,
                format=logformat,
                datefmt=dateformat)
            logging.error("Unable to write to log file " + logpath)
            returnValue = False
        return returnValue

    @staticmethod
    def fileIsPDF(filedata):
        """Check if a file is or isnt a PDF

        Args:
        filename: string, name of the file to check
        Returns:
        boolean: True = is a PDF, False = not a PDF.
        """
        p = subprocess.Popen(["file", '-'], stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE, stderr=subprocess.PIPE)
        output = p.communicate(filedata)[0]
        logging.debug("File output was: " + output)
        return "PDF document" in output

    @staticmethod
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    @staticmethod
    def getPath():
        return os.environ["PATH"]

    @staticmethod
    def which(program):
        shellpath = Utils.getPath()
        logging.debug("Checking path is %s" % shellpath)
        for path in shellpath.split(os.pathsep):
            exe_file = os.path.join(path, program)
            if Utils.is_exe(exe_file):
                return exe_file
        return None

    @staticmethod
    def GetLPID(default='lp', alternative='cups', useFiles=True,
                blacklistedGroups=None,
                useFilesOnly=False):
        if blacklistedGroups is None:
            blacklistedGroups = ['adm', 'wheel', 'root']
        blacklistedGroupIds = []
        for group in blacklistedGroups:
            try:
                blacklistedGroupIds.append(grp.getgrnam(group).gr_gid)
            except Exception:
                logging.debug("Group " + group + " not found")

        if useFiles:
            # check files in order
            for cupsConfigFile in ['/var/log/cups/access_log',
                                   '/etc/cups/ppd',
                                   '/usr/local/etc/cups/ppd']:
                if os.path.exists(cupsConfigFile):
                    configGid = os.stat(cupsConfigFile).st_gid
                    if configGid not in blacklistedGroupIds:
                        return configGid
                    else:
                        logging.debug(
                            "Group " +
                            group +
                            " excluded as blacklisted")

        if useFilesOnly:
            return None

        # try lp first, then cups
        lpgrp = None
        try:
            lpgrp = grp.getgrnam(default)
        except Exception:
            try:
                lpgrp = grp.getgrnam(alternative)
            except Exception:
                pass
        if lpgrp is None:
            return None
        else:
            return lpgrp.gr_gid

    @staticmethod
    def ShowVersion(CCPVersion):
        if len(sys.argv) == 2 and sys.argv[1] == 'version':
            print "CUPS Cloud Print Version " + CCPVersion
            sys.exit(0)
        return False

    @staticmethod
    def ReadFile(pathname):
        """Read contents of a file and return content.

        Args:
          pathname: string, (path)name of file.
        Returns:
          string: contents of file.
        """
        try:
            f = open(pathname, 'rb')
            s = f.read()
            return s
        except IOError as e:
            print 'ERROR: Error opening %s\n%s', pathname, e
            return None

    @staticmethod
    def WriteFile(file_name, data):
        """Write contents of data to a file_name.

        Args:
          file_name: string, (path)name of file.
          data: string, contents to write to file.
        Returns:
          boolean: True = success, False = errors.
        """
        status = True

        try:
            f = open(file_name, 'wb')
            f.write(data)
            f.close()
        except IOError:
            status = False

        return status

    @staticmethod
    def Base64Encode(data, jobtype):
        """Convert a file to a base64 encoded file.

        Args:
          pathname: data to base64 encode
          jobtype: job type being encoded - pdf, jpg etc
        Returns:
          string, base64 encoded string.
        For more info on data urls, see:
          http://en.wikipedia.org/wiki/Data_URI_scheme
        """
        # Convert binary data to base64 encoded data.
        mimetype = Utils._MIMETYPES_JOBTYPES['other']
        if jobtype in Utils._MIMETYPES_JOBTYPES:
            mimetype = Utils._MIMETYPES_JOBTYPES[jobtype]
        header = 'data:%s;base64,' % mimetype
        return header + base64.b64encode(data)

    @staticmethod
    def GetLanguage(locale):
        language = 'en'
        if len(locale) < 1 or locale[0] is None:
            return language
        defaultlocale = locale[0]
        language = defaultlocale
        if '_' in language:
            language = language.split("_")[0]
        return language

    @staticmethod
    def GetDefaultPaperType(locale):
        defaultpapertype = "Letter"
        if len(locale) < 1 or locale[0] is None:
            return defaultpapertype
        if len(locale[0].split('_')) > 1 and locale[0].split('_')[1] not in Utils._LETTER_COUNTRIES:
            defaultpapertype = "A4"
        return defaultpapertype

    @staticmethod
    def GetWindowSize(winsize=None):
        """Gets window height and width.

        Gets window (aka terminal, console) height and width using IOCtl Get WINdow SiZe
        method.

        Returns:
            The tuple (height, width) of the window as integers, or None if the
            windows size isn't available.
        """
        try:
            bytes = struct.pack('HHHH', 0, 0, 0, 0)
            if winsize is None:
                winsize = fcntl.ioctl(sys.stdout.fileno(), termios.TIOCGWINSZ, bytes)
            height, width = struct.unpack('HHHH', winsize)[:2]
        except Exception:
            return None

        if height > 0 and width > 0:
            return height, width
        return None

    @staticmethod
    def GetTempFileName():
        tempdir = '/tmp/'
        if tempfile.tempdir:
            tempdir = tempfile.tempdir
        if 'TMPDIR' in os.environ:
            tempdir = os.environ['TMPDIR']
        return os.path.join(tempdir, (''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(32))))



#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

if __name__ == '__main__':  # pragma: no cover

    import sys
    import cups
    import subprocess
    import os
    import json
    import logging
    import urllib
    from auth import Auth
    from ccputils import Utils
    from printermanager import PrinterManager
    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    if not os.path.exists("/etc/cloudprint.conf"):
        sys.stderr.write(
            "Config is invalid or missing, not running on fresh install\n")
        logging.warning("Upgrade tried to run on fresh install")
        sys.exit(0)

    requestors, storage = Auth.SetupAuth(False)
    if not requestors:
        sys.stderr.write("Config is invalid or missing\n")
        logging.error("Upgrade tried to run with invalid config")
        sys.exit(0)
    printer_manager = PrinterManager(requestors)

    logging.info("Upgrading to " + CCPVersion)

    try:
        connection = cups.Connection()
    except Exception as e:
        sys.stderr.write("Could not connect to CUPS: " + e.message + "\n")
        sys.exit(0)
    cupsprinters = connection.getPrinters()

    if os.path.exists(Auth.config):
        Utils.FixFilePermissions(Auth.config)

        try:
            content_file = open(Auth.config, 'r')
            content = content_file.read()
            data = json.loads(content)
        except Exception as e:
            sys.stderr.write(
                "Unable to read config file: " +
                e.message +
                "\n\n")
            sys.exit(0)

    else:
        errormessage = "\nRun: /usr/share/cloudprint-cups/"
        errormessage += "setupcloudprint.py to"
        errormessage += " setup your Google Credentials"
        errormessage += " and add your printers to CUPS\n\n"
        sys.stderr.write(errormessage)
        sys.exit(0)

    from ccputils import Utils
    if Utils.which('lpadmin') is None:
        errormessage = "lpadmin command not found"
        errormessage += ", you may need to run this script as root\n"
        sys.stderr.write(errormessage)
        sys.exit(1)

    try:
        print "Fetching list of available ppds..."
        allppds = connection.getPPDs()
        print "List retrieved successfully"
    except Exception as e:
        sys.stderr.write("Error connecting to CUPS: " + str(e) + "\n")
        sys.exit(1)

    for device in cupsprinters:
        try:
            if (cupsprinters[device]["device-uri"].find(Utils._OLD_PROTOCOL) == 0)\
                    or (cupsprinters[device]["device-uri"].find(Utils._PROTOCOL) == 0):
                account, printername, printerid, formatid = \
                    printer_manager.parseLegacyURI(
                        cupsprinters[device]["device-uri"],
                        requestors)
                if formatid != PrinterManager.URIFormatLatest:
                    # not latest format, needs upgrading
                    updatingmessage = "Updating "
                    updatingmessage += cupsprinters[device]["printer-info"]
                    updatingmessage += " with new id uri format"
                    print updatingmessage
                    tempprinter = None
                    printerid, requestor = printer_manager.getPrinterIDByDetails(
                        account, printername, printerid)
                    if printerid is not None:
                        tempprinter = printer_manager.getPrinter(
                            printerid,
                            urllib.unquote(account))
                    if tempprinter is not None:
                        cupsprinters[device]["device-uri"] = tempprinter.getURI()
                        p = subprocess.Popen(
                            ["lpadmin",
                             "-p",
                             cupsprinters[device]["printer-info"].lstrip('-'),
                                "-v",
                                tempprinter.getURI()],
                            stderr=subprocess.PIPE,
                            stdout=subprocess.PIPE)
                        output = p.communicate()[0]
                        result = p.returncode
                        sys.stderr.write(output)
                    else:
                        errormessage = cupsprinters[device]["printer-info"]
                        errormessage += " not found, "
                        errormessage += "you should delete and "
                        errormessage += "re-add this printer"
                        print errormessage
                        continue
                else:
                    print "Updating " + cupsprinters[device]["printer-info"]

                ppdid = 'MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL:' + \
                    cupsprinters[device]["device-uri"]

                # just needs updating
                printerppdname = None
                for ppd in allppds:
                    if allppds[ppd]['ppd-device-id'] == ppdid:
                        printerppdname = ppd
                if printerppdname is not None:
                    p = subprocess.Popen(
                        ["lpadmin",
                         "-p",
                         cupsprinters[device]["printer-info"].lstrip('-'),
                            "-m",
                            printerppdname.lstrip('-')],
                        stderr=subprocess.PIPE,
                        stdout=subprocess.PIPE)
                    output = p.communicate()[0]
                    result = p.returncode
                    sys.stderr.write(output)
                else:
                    errormessage = cupsprinters[device]["printer-info"]
                    errormessage += " not found, you should delete and"
                    errormessage += " re-add this printer"
                    print errormessage
        except Exception as e:
            sys.stderr.write("Error connecting to CUPS: " + str(e) + "\n")



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A service account credentials class.

This credentials class is implemented on top of rsa library.
"""

import base64
import time

from pyasn1.codec.ber import decoder
from pyasn1_modules.rfc5208 import PrivateKeyInfo
import rsa

from oauth2client import GOOGLE_REVOKE_URI
from oauth2client import GOOGLE_TOKEN_URI
from oauth2client._helpers import _json_encode
from oauth2client._helpers import _to_bytes
from oauth2client._helpers import _urlsafe_b64encode
from oauth2client import util
from oauth2client.client import AssertionCredentials


class _ServiceAccountCredentials(AssertionCredentials):
    """Class representing a service account (signed JWT) credential."""

    MAX_TOKEN_LIFETIME_SECS = 3600  # 1 hour in seconds

    def __init__(self, service_account_id, service_account_email,
                 private_key_id, private_key_pkcs8_text, scopes,
                 user_agent=None, token_uri=GOOGLE_TOKEN_URI,
                 revoke_uri=GOOGLE_REVOKE_URI, **kwargs):

        super(_ServiceAccountCredentials, self).__init__(
            None, user_agent=user_agent, token_uri=token_uri,
            revoke_uri=revoke_uri)

        self._service_account_id = service_account_id
        self._service_account_email = service_account_email
        self._private_key_id = private_key_id
        self._private_key = _get_private_key(private_key_pkcs8_text)
        self._private_key_pkcs8_text = private_key_pkcs8_text
        self._scopes = util.scopes_to_string(scopes)
        self._user_agent = user_agent
        self._token_uri = token_uri
        self._revoke_uri = revoke_uri
        self._kwargs = kwargs

    def _generate_assertion(self):
        """Generate the assertion that will be used in the request."""

        header = {
            'alg': 'RS256',
            'typ': 'JWT',
            'kid': self._private_key_id
        }

        now = int(time.time())
        payload = {
            'aud': self._token_uri,
            'scope': self._scopes,
            'iat': now,
            'exp': now + _ServiceAccountCredentials.MAX_TOKEN_LIFETIME_SECS,
            'iss': self._service_account_email
        }
        payload.update(self._kwargs)

        first_segment = _urlsafe_b64encode(_json_encode(header))
        second_segment = _urlsafe_b64encode(_json_encode(payload))
        assertion_input = first_segment + b'.' + second_segment

        # Sign the assertion.
        rsa_bytes = rsa.pkcs1.sign(assertion_input, self._private_key,
                                   'SHA-256')
        signature = base64.urlsafe_b64encode(rsa_bytes).rstrip(b'=')

        return assertion_input + b'.' + signature

    def sign_blob(self, blob):
        # Ensure that it is bytes
        blob = _to_bytes(blob, encoding='utf-8')
        return (self._private_key_id,
                rsa.pkcs1.sign(blob, self._private_key, 'SHA-256'))

    @property
    def service_account_email(self):
        return self._service_account_email

    @property
    def serialization_data(self):
        return {
            'type': 'service_account',
            'client_id': self._service_account_id,
            'client_email': self._service_account_email,
            'private_key_id': self._private_key_id,
            'private_key': self._private_key_pkcs8_text
        }

    def create_scoped_required(self):
        return not self._scopes

    def create_scoped(self, scopes):
        return _ServiceAccountCredentials(self._service_account_id,
                                          self._service_account_email,
                                          self._private_key_id,
                                          self._private_key_pkcs8_text,
                                          scopes,
                                          user_agent=self._user_agent,
                                          token_uri=self._token_uri,
                                          revoke_uri=self._revoke_uri,
                                          **self._kwargs)


def _get_private_key(private_key_pkcs8_text):
    """Get an RSA private key object from a pkcs8 representation."""
    private_key_pkcs8_text = _to_bytes(private_key_pkcs8_text)
    der = rsa.pem.load_pem(private_key_pkcs8_text, 'PRIVATE KEY')
    asn1_private_key, _ = decoder.decode(der, asn1Spec=PrivateKeyInfo())
    return rsa.PrivateKey.load_pkcs1(
        asn1_private_key.getComponentByName('privateKey').asOctets(),
        format='DER')



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import cups
import urllib
import logging
import sys
import subprocess
import os
import pytest
import shutil
sys.path.insert(0, ".")
from printermanager import PrinterManager
from mockrequestor import MockRequestor
from ccputils import Utils

global printers, printerManagerInstance

testCapabilities1 = {'printer':
                     {'color':
                      {
                          'option':
                          [
                              {'is_default': True, 'vendor_id': '2', 'type': 'STANDARD_COLOR'},
                              {'vendor_id': '1', 'type': 'STANDARD_MONOCHROME'}
                          ]
                     },
                     'media_size':
                     {
                          'option':
                          [
                                {'vendor_id': '1', 'height_microns': 279400, 'width_microns': 215900, 'name': 'NA_LETTER', 'custom_display_name': 'Letter'},
                                {'vendor_id': '3', 'height_microns': 431800, 'width_microns': 279400, 'name': 'NA_LEDGER', 'custom_display_name': 'Tabloid'},
                                {'vendor_id': '5', 'height_microns': 355600, 'width_microns': 215900, 'name': 'NA_LEGAL', 'custom_display_name': 'Legal'},
                                {'vendor_id': '6', 'height_microns': 215900, 'width_microns': 139700, 'name': 'NA_INVOICE', 'custom_display_name': 'Statement'},
                                {'vendor_id': '7', 'height_microns': 266700, 'width_microns': 184100, 'name': 'NA_EXECUTIVE', 'custom_display_name': 'Executive'},
                                {'vendor_id': '8', 'height_microns': 420000, 'width_microns': 297000, 'name': 'ISO_A3', 'custom_display_name': 'A3'},
                                {'width_microns': 210000, 'vendor_id': '9', 'is_default': True, 'height_microns': 297000, 'custom_display_name': 'A4', 'name': 'ISO_A4'},
                                {'vendor_id': '11', 'height_microns': 210000, 'width_microns': 148000, 'name': 'ISO_A5', 'custom_display_name': 'A5'},
                                {'vendor_id': '12', 'height_microns': 364000, 'width_microns': 257000, 'name': 'JIS_B4', 'custom_display_name': 'B4 (JIS)'},
                                {'vendor_id': '13', 'height_microns': 257000, 'width_microns': 182000, 'name': 'JIS_B5', 'custom_display_name': 'B5 (JIS)'}
                          ]
                     },
                     'page_orientation':
                     {
                          'option':
                          [
                              {'is_default': True, 'type': 'PORTRAIT', 'display_name' : 'portrait'},
                              {'type': 'LANDSCAPE', 'custom_display_name' : 'landscape'},
                              {'type': 'AUTO'}
                          ]
                     },
                     'something':
                     {
                          'option':
                          [
                              {'is_default': True, 'type': 'one'},
                              {'type': 'testval'}
                          ]
                     },
                     'test-reserved-word':
                     {
                          'option':
                          [
                              {'is_default': True, 'type': 'Resolution'},
                              {'type': 'testtwo'}
                          ]
                     },
                     'supported_content_type':
                     [
                          {'content_type': 'application/pdf'}
                     ],
                     'dpi':
                     {
                          'option':
                          [
                              {'is_default': True, 'horizontal_dpi': 600, 'vertical_dpi': 600 },
                              {'horizontal_dpi': 800, 'vertical_dpi': 800 }
                          ]
                     },
                     'empty_options':
                     {
                          'option':
                          [
                          ]
                     }
                     },
                     'version': '1.0'
                    }

def setup_function(function):
    # setup mock requestors
    global printers, printerManagerInstance

    mockRequestorInstance = MockRequestor()
    mockRequestorInstance.setAccount('testaccount2@gmail.com')
    mockRequestorInstance.printers = [{'name': 'Save to Google Drive',
                                       'id': '__test_save_docs',
                                       'capabilities': testCapabilities1},
                                      {'name': 'Save to Google Drive 2',
                                       'displayName': 'Save to Google Drive 2 DisplayName',
                                       'id': '__test_save_docs_2',
                                       'capabilities' : {'printer' : {}}}
                                      ]

    printerManagerInstance = PrinterManager(mockRequestorInstance)
    printers = printerManagerInstance.getPrinters()


def teardown_function(function):
    global requestors
    requestors = None
    logging.shutdown()
    reload(logging)


def test_getAccount():
    global printers
    for printer in printers:
        assert printer.getAccount() == "testaccount2@gmail.com"


def test_getRequestor():
    global printers
    for printer in printers:
        requestor = printer.getRequestor()
        assert requestor.__class__.__name__ == "MockRequestor"
        assert requestor.getAccount() == 'testaccount2@gmail.com'


def test_getMimeBoundary():
    global printers
    for printer in printers:
        assert printer._getMimeBoundary() != 'test_boundry'
        assert len(printer._getMimeBoundary()) > 30
        assert len(printer._getMimeBoundary()) < 50

        printer._mime_boundary = 'test_boundry'
        assert printer._getMimeBoundary() == 'test_boundry'


def test_getCapabilitiesItems():
    global printers
    printer = printers[0]
    correctCapabilities = testCapabilities1
    assert printer._fields['capabilities'] == correctCapabilities
    assert printer._fields['capabilities'] == printer['capabilities']
    del printer._fields['capabilities']
    assert 'capabilities' not in printer._fields
    assert printer['capabilities'] == correctCapabilities
    assert printer._fields['capabilities'] == printer['capabilities']


def test_getCapabilitiesItemsMissing():
    global printers
    printer = printers[1]
    assert 'capabilities' in printer._fields
    assert printer['capabilities'] == {'printer' : {}}


def test_contains():
    global printers
    for printer in printers:
        assert 'testvalue' not in printer
        printer._fields['testvalue'] = 'test'
        assert 'testvalue' in printer
        del printer._fields['testvalue']
        assert 'testvalue' not in printer


def test_fetchDetails():
    global printers
    assert printers[0]._fetchDetails() == {'name': 'Save to Google Drive',
                                           'id': '__test_save_docs',
                                           'capabilities': testCapabilities1}
    assert printers[1]._fetchDetails() == {'displayName': 'Save to Google Drive 2 DisplayName',
                                           'id': '__test_save_docs_2', 'name': 'Save to Google Drive 2',
                                           'capabilities': {'printer' : {}}}


def test_getURI():
    global printers
    assert printers[0].getURI() == Utils._PROTOCOL + "testaccount2%40gmail.com/__test_save_docs"
    assert printers[1].getURI() == Utils._PROTOCOL + "testaccount2%40gmail.com/__test_save_docs_2"


def test_getDisplayName():
    global printers
    assert printers[0].getDisplayName() == "Save to Google Drive"
    assert printers[1].getDisplayName() == "Save to Google Drive 2 DisplayName"


def test_getListDescription():
    global printers
    assert printers[0].getListDescription() == "Save to Google Drive - " + Utils._PROTOCOL + \
        "testaccount2%40gmail.com/__test_save_docs - testaccount2@gmail.com"
    assert printers[1].getListDescription() == "Save to Google Drive 2 DisplayName - " + \
        Utils._PROTOCOL + "testaccount2%40gmail.com/__test_save_docs_2 - testaccount2@gmail.com"


def test_getLocation():
    global printers
    assert printers[0].getLocation() == ''
    printers[0]._fields['tags'] = ['novalue', 'name=value', 'location=test-location']
    assert printers[0].getLocation() == 'test-location'


def test_getCUPSBackendDescription():
    global printers
    assert printers[0].getCUPSBackendDescription() == 'network ' + Utils._PROTOCOL + \
        'testaccount2%40gmail.com/__test_save_docs "Save to Google Drive" "Save to Google Drive" "MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL:' + \
        Utils._PROTOCOL + 'testaccount2%40gmail.com/__test_save_docs"'
    assert printers[1].getCUPSBackendDescription() == 'network ' + Utils._PROTOCOL + \
        'testaccount2%40gmail.com/__test_save_docs_2 "Save to Google Drive 2 DisplayName" "Save to Google Drive 2 DisplayName" "MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL:' + \
        Utils._PROTOCOL + 'testaccount2%40gmail.com/__test_save_docs_2"'
    printers[0]._fields['tags'] = ['novalue', 'name=value', 'location=test-location']
    assert printers[0].getCUPSBackendDescription() == 'network ' + Utils._PROTOCOL + 'testaccount2%40gmail.com/__test_save_docs "Save to Google Drive" "Save to Google Drive @ test-location" "MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL:' + \
        Utils._PROTOCOL + 'testaccount2%40gmail.com/__test_save_docs" "test-location"'


def test_getCUPSDriverDescription():
    global printers
    assert printers[0].getCUPSDriverDescription() == '"cupscloudprint:testaccount2%40gmail.com:__test_save_docs.ppd" en "Google" "Save to Google Drive (testaccount2@gmail.com)" "MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL:' + \
        Utils._PROTOCOL + 'testaccount2%40gmail.com/__test_save_docs"'
    assert printers[1].getCUPSDriverDescription() == '"cupscloudprint:testaccount2%40gmail.com:__test_save_docs_2.ppd" en "Google" "Save to Google Drive 2 DisplayName (testaccount2@gmail.com)" "MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL:' + \
        Utils._PROTOCOL + 'testaccount2%40gmail.com/__test_save_docs_2"'


def test_getPPDName():
    global printers
    assert printers[0].getPPDName(
    ) == "cupscloudprint:testaccount2%40gmail.com:__test_save_docs.ppd"
    assert printers[1].getPPDName(
    ) == "cupscloudprint:testaccount2%40gmail.com:__test_save_docs_2.ppd"


def test_generatePPD():
    global printers
    for printer in printers:
        ppddata = printer.generatePPD()
        assert isinstance(ppddata, basestring)

        # test ppd data is valid
        tempfile = open('/tmp/.ppdfile', 'w')
        tempfile.write(ppddata)
        tempfile.close()

        p = subprocess.Popen(['cupstestppd', '/tmp/.ppdfile'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        testdata = p.communicate()[0]
        print testdata
        os.unlink('/tmp/.ppdfile')
        assert p.returncode == 0


def test_sanitizeText():
    global printers
    for checkReserved in [False, True]:
        assert printers[0]._sanitizeText("", checkReserved) == ""
        assert printers[0]._sanitizeText("TESTSTRING", checkReserved) == "TESTSTRING"
        assert printers[0]._sanitizeText("TEST:; STRING /2", checkReserved) == "TEST___STRING_-2"
        assert printers[0]._sanitizeText("TEST:; STRING /2", checkReserved) == "TEST___STRING_-2"

    assert printers[0]._sanitizeText("Duplex") == "Duplex"
    assert printers[0]._sanitizeText("Duplex", True) == "GCP_Duplex"

    # ensure strips newlines
    assert printers[0]._sanitizeText("Test\ntest2") == "Testtest2"
    assert printers[0]._sanitizeText("Test\r\ntest2") == "Testtest2"


def test_getInternalName():
    global printers
    printerItem = printers[0]

    internalCapabilityTests = []
    # load test file and try all those
    for filelineno, line in enumerate(open('testing/testfiles/capabilitylist')):
        internalCapabilityTests.append({'name': line.decode("utf-8")})

    for internalTest in internalCapabilityTests:
        assert printerItem._getInternalName(
            internalTest, 'capability') not in printerItem._RESERVED_CAPABILITY_WORDS
        assert not printerItem._getInternalName(internalTest, 'capability').startswith(
            printerItem._RESERVED_CAPABILITY_PREFIXES)
        assert ':' not in printerItem._getInternalName(internalTest, 'capability')
        assert ' ' not in printerItem._getInternalName(internalTest, 'capability')
        assert len(printerItem._getInternalName(internalTest, 'capability')) <= 30
        assert len(printerItem._getInternalName(internalTest, 'capability')) >= 1

    for internalTest in internalCapabilityTests:
        for capabilityName in ["psk:JobDuplexAllDocumentsContiguously", "other", "psk:PageOrientation", "cupsFilter"]:
            assert printerItem._getInternalName(
                internalTest, 'option', capabilityName) not in printerItem._RESERVED_CAPABILITY_WORDS
            assert not printerItem._getInternalName(internalTest, 'option', capabilityName).startswith(
                printerItem._RESERVED_CAPABILITY_PREFIXES)
            assert ':' not in printerItem._getInternalName(internalTest, 'option', capabilityName)
            assert ' ' not in printerItem._getInternalName(internalTest, 'option')
            assert len(printerItem._getInternalName(internalTest, 'option', capabilityName)) <= 30
            assert len(printerItem._getInternalName(internalTest, 'option', capabilityName)) >= 1

    # ensure fixed mappings works correctly
    for fixedCapabilityName in printerItem._FIXED_CAPABILITY_MAPPINGS:
        assert printerItem._getInternalName(
            {'name': fixedCapabilityName}, 'capability') == printerItem._FIXED_CAPABILITY_MAPPINGS[fixedCapabilityName]

    for fixedCapabilityName in printerItem._FIXED_OPTION_MAPPINGS:
        for fixedOptionName in printerItem._FIXED_OPTION_MAPPINGS[fixedCapabilityName]:
            assert printerItem._getInternalName({'name': fixedOptionName}, 'option', fixedCapabilityName) == printerItem._FIXED_OPTION_MAPPINGS[
                fixedCapabilityName][fixedOptionName]


def test_encodeMultiPart():
    global printers
    assert isinstance(printers[0]._encodeMultiPart([('test', 'testvalue')]), basestring)
    assert 'testvalue' in printers[0]._encodeMultiPart([('test', 'testvalue')])
    assert 'Content-Disposition: form-data; name="test"' in printers[
        0]._encodeMultiPart([('test', 'testvalue')])


def test_getOverrideCapabilities():
    global printers
    printerItem = printers[0]
    assert printerItem._getOverrideCapabilities("") == {}
    assert printerItem._getOverrideCapabilities("landscape") == {'Orientation': 'Landscape'}
    assert printerItem._getOverrideCapabilities("nolandscape") == {'Orientation': 'Landscape'}
    assert printerItem._getOverrideCapabilities("test=one") == {'test': 'one'}
    assert printerItem._getOverrideCapabilities("test=one anothertest=two") == {
        'test': 'one', 'anothertest': 'two'}
    assert printerItem._getOverrideCapabilities("test=one anothertest=two Orientation=yes") == {
        'test': 'one', 'anothertest': 'two'}


def test_GetCapabilitiesDict():
    global printers
    printerItem = printers[0]
    assert printerItem._getCapabilitiesDict({}, {}, {}) == {'print': {}, 'version': '1.0'}
    assert printerItem._getCapabilitiesDict([{'name': 'test'}], {}, {}) == {'print': {}, 'version': '1.0'}
    assert printerItem._getCapabilitiesDict([{'name': 'Default' + 'test123', 'value': 'STANDARD_MONOCHROME'}, { 'name' : 'Defaulttestname', 'value': 'test1' }],
                                            {'test123':
                                                        {
                                                            'option':
                                                            [
                                                                {'is_default': True, 'vendor_id': '2', 'type': 'STANDARD_COLOR'},
                                                                {'vendor_id': '1', 'type': 'STANDARD_MONOCHROME'}
                                                            ]
                                                        },
                                             'testname':
                                                        {
                                                            'option':
                                                            [
                                                                {'is_default': True, 'vendor_id': '2', 'name': 'test1', "test_param_1" : "test123", "test_param_2" : "test456"},
                                                                {'vendor_id': '1', 'name': 'test2', "test_param_1" : "test1", "test_param_2" : "test2"}
                                                            ]
                                                        },
                                             'testnooption':
                                                        {
                                                            'option':
                                                            [
                                                                {'is_default': True, 'vendor_id': '2', 'horizontal_dpi': 300, 'vertical_dpi' : 300, 'display_name': 'test' },
                                                                {'vendor_id': '2', 'horizontal_dpi': 400, 'vertical_dpi' : 400, 'display_name': 'test123' }
                                                            ]
                                                        }
                                                            
                                                            }, {'test123': 'STANDARD_MONOCHROME'}) == {
                                                                                                    "version": "1.0",
                                                                                                        "print": {
                                                                                                            "testname":{"vendor_id":"2", "test_param_1" : "test123", "test_param_2" : "test456"},
                                                                                                            "test123": {"type": "STANDARD_MONOCHROME"}
                                                                                                        }
                                                                                                    }


def test_attrListToArray():
    global printers
    assert len(list(printers[0]._attrListToArray({}))) == 0


def test_getCapabilities():
    global printers, printerManagerInstance
    printer = printers[1]
    connection = cups.Connection()

    # get test ppd
    ppdid = 'MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL'
    ppds = connection.getPPDs(ppd_device_id=ppdid)
    printerppdname, printerppd = ppds.popitem()

    assert printerManagerInstance.addPrinter(
        printer['name'],
        printer,
        connection,
        printerppdname) is not None
    emptyoptions = printer._getCapabilities(
        printerManagerInstance.sanitizePrinterName(printer['name']), "landscape")
    assert isinstance(emptyoptions, dict)
    assert isinstance(emptyoptions['print'], dict)
    assert len(emptyoptions['print']) == 0
    connection.deletePrinter(printerManagerInstance.sanitizePrinterName(printer['name']))


def test_submitJob():
    global printers, printerManagerInstance
    printer = printers[0]
    connection = cups.Connection()
    testprintername = printerManagerInstance.sanitizePrinterName(printer['name'])

    # get test ppd
    ppdid = 'MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL'
    ppds = connection.getPPDs(ppd_device_id=ppdid)
    printerppdname, printerppd = ppds.popitem()

    assert printerManagerInstance.addPrinter(
        printer['name'],
        printer,
        connection,
        printerppdname) is not None

    # test submitting job
    assert printer.submitJob(
        'pdf',
        'testing/testfiles/Test Page.pdf',
        open('testing/testfiles/Test Page.pdf').read(),
        'Test Page',
        testprintername) == True
    assert printer.submitJob(
        'pdf',
        'testing/testfiles/Test Page Doesnt Exist.pdf',
        '',
        'Test Page',
        testprintername) == False
    assert printer.submitJob(
        'pdf',
        'testing/testfiles/Test Page Corrupt.pdf',
        open('testing/testfiles/Test Page Corrupt.pdf').read(),
        'Test Page',
        testprintername, 'landscape') == False

    # copy for rotating
    tmpDir = os.getenv('TMPDIR')
    if not tmpDir:
        tmpDir = "/tmp"
    tmpFile = os.path.join(tmpDir, 'Test Page.pdf')
    shutil.copy('testing/testfiles/Test Page.pdf', tmpFile)

    # test submitting job with rotate
    assert printer.submitJob(
        'pdf',
        tmpFile,
        open(tmpFile).read(),
        'Test Page',
        testprintername,
        "landscape") == True
    assert printer.submitJob(
        'pdf',
        tmpFile,
        open(tmpFile).read(),
        'Test Page',
        testprintername,
        "nolandscape") == True

    os.unlink(tmpFile)

    # test submitting job with no filename
    assert printer.submitJob(
        'pdf',
        '',
        'data',
        '',
        testprintername) == True

    # test submitting job with no name
    assert printer.submitJob(
        'pdf',
        'testing/testfiles/Test Page.pdf',
        open('testing/testfiles/Test Page.pdf').read(),
        '',
        testprintername) == True
    assert printer.submitJob(
        'pdf',
        'testing/testfiles/Test Page Doesnt Exist.pdf',
        '',
        '',
        testprintername) == False

    # png
    assert printer.submitJob(
        'png',
        'testing/testfiles/Test Page.png',
        open('testing/testfiles/Test Page.png').read(),
        'Test Page',
        testprintername) == True
    assert printer.submitJob(
        'png',
        'testing/testfiles/Test Page Doesnt Exist.png',
        '',
        'Test Page',
        testprintername) == False

    # ps
    assert printer.submitJob(
        'ps',
        'testing/testfiles/Test Page.ps',
        open('testing/testfiles/Test Page.ps').read(),
        'Test Page',
        testprintername) == False
    assert printer.submitJob(
        'ps',
        'testing/testfiles/Test Page Doesnt Exist.ps',
        '',
        'Test Page',
        testprintername) == False

    # test failure of print job
    assert printer.submitJob(
        'pdf',
        'testing/testfiles/Test Page.pdf',
        open('testing/testfiles/Test Page.pdf').read(),
        'FAIL PAGE',
        testprintername) == False

    # test failure of print job with exception
    assert printer.submitJob(
        'pdf',
        'testing/testfiles/Test Page.pdf',
        open('testing/testfiles/Test Page.pdf').read(),
        'TEST PAGE WITH EXCEPTION',
        testprintername) == False

    # delete test printer
    connection.deletePrinter(testprintername)


@pytest.mark.skipif(
    os.getuid() == 0,
    reason="will only pass if running tests as non-root user")
def test_submitJobFileCreationFails():
    global printers, printerManagerInstance
    printer = printers[0]
    connection = cups.Connection()
    testprintername = printerManagerInstance.sanitizePrinterName(printer['name'])

    # get test ppd
    ppdid = 'MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL'
    ppds = connection.getPPDs(ppd_device_id=ppdid)
    printerppdname, printerppd = ppds.popitem()

    assert printerManagerInstance.addPrinter(
        printer['name'],
        printer,
        connection,
        printerppdname) is not None

    # delete test printer
    connection.deletePrinter(testprintername)



#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2013 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import urllib
import sys


def doList(sys, printer_manager):
    """Lists Google Cloud Print printers."""
    printers = printer_manager.getPrinters()
    if printers is None:
        sys.stderr.write("ERROR: No Printers Found\n")
        sys.exit(1)
    for printer in printers:
        print printer.getCUPSDriverDescription()
    sys.exit(0)


def doCat():
    """Prints a PPD to stdout, per argv arguments."""
    try:
        drivername, accountName, printerId = sys.argv[2].split(':', 2)
        if drivername != 'cupscloudprint' or not printerId.endswith('.ppd'):
            raise Exception('Bad PPD name format')

        accountName = urllib.unquote(accountName)
        printerId = printerId[:-4]

    except Exception:
        sys.stderr.write("ERROR: PPD name '%s' is invalid\n" % sys.argv[2])
        sys.exit(1)

    printer = printer_manager.getPrinter(printerId, accountName)

    if printer is None:
        sys.stderr.write("ERROR: PPD %s not found\n" % sys.argv[2])
        sys.exit(1)

    print printer.generatePPD()

    sys.exit(0)


def showUsage():
    sys.stderr.write("ERROR: Usage: %s [list|version|cat drivername]\n" % sys.argv[0])
    sys.exit(1)

if __name__ == '__main__':  # pragma: no cover

    libpaths = [
        "/Library/cloudprint-cups/",
        "/usr/local/share/cloudprint-cups/",
        "/usr/share/cloudprint-cups"
    ]
    addedPath = False
    for libpath in libpaths:
        if os.path.exists(libpath):
            sys.path.insert(0, libpath)
            addedPath = True
            break

    if not addedPath:
        sys.stderr.write("ERROR: Could not find any valid path for python files\n")
        sys.exit(1)

    from auth import Auth
    from printermanager import PrinterManager
    from ccputils import Utils

    if len(sys.argv) < 2:
        showUsage()

    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    requestors, storage = Auth.SetupAuth(False)
    if not requestors:
        sys.stderr.write("ERROR: config is invalid or missing\n")
        logging.error("backend tried to run with invalid config")
        sys.exit(1)

    printer_manager = PrinterManager(requestors)

    if sys.argv[1] == 'list':
        doList(sys, printer_manager)

    elif sys.argv[1] == 'cat':
        if len(sys.argv) == 2 or sys.argv[2] == "":
            showUsage()
        doCat()

    showUsage()



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import logging
import sys
import pytest
import struct
sys.path.insert(0, ".")

from ccputils import Utils

def teardown_function(function):
    logging.shutdown()
    reload(logging)

def test_SetupLogging():
    testLogFile = '/tmp/testccp.log'
    assert os.path.exists(testLogFile) == False
    assert Utils.SetupLogging(testLogFile) == True
    logging.error('test_setupLogging error test')
    assert os.path.exists(testLogFile) == True
    os.unlink(testLogFile)

def test_SetupLoggingDefault():
    testLogFile = '/tmp/testccp.log'
    assert os.path.exists(testLogFile) == False
    Utils.logpath = testLogFile
    assert Utils.SetupLogging() == True
    logging.error('test_setupLogging error test')
    assert os.path.exists(testLogFile) == True
    os.unlink(testLogFile)

def test_SetupLoggingFails():
    testLogFile = '/tmp/dirthatdoesntexist/testccp.log'
    assert os.path.exists(testLogFile) == False
    assert Utils.SetupLogging(testLogFile) == False
    assert os.path.exists(testLogFile) == False

def test_fileIsPDFFails():
    assert Utils.fileIsPDF(open('testing/testfiles/NotPdf.txt').read()) == False

def test_fileIsPDFSucceeds():
    assert Utils.fileIsPDF(open('testing/testfiles/Test Page.pdf').read(128)) == True

def test_fileIsPDFErrors():
    assert Utils.fileIsPDF("testdata") == False

def test_whichFails():
    assert Utils.which('dsaph9oaghd9ahdsadsadsadsadasd') is None

def test_whichSucceeds():
    assert Utils.which(
        'bash') in (
        '/bin/bash',
        '/usr/bin/bash',
        '/usr/sbin/bash')

def test_isExeSucceeds():
    if os.path.exists('/usr/bin/sh'):
        assert Utils.is_exe("/usr/bin/sh") == True
    else:
        assert Utils.is_exe("/bin/sh") == True


def test_isExeFails():
    assert Utils.is_exe("/etc/passwd") == False


def test_getLPID():
    assert int(Utils.GetLPID()) > 0
    assert Utils.GetLPID() is not None

    import grp

    workingPrintGroupName = 'lp'
    try:
        grp.getgrnam(workingPrintGroupName)
    except:
        workingPrintGroupName = 'cups'
        pass

    assert Utils.GetLPID('brokendefault', 'brokenalternative', False) is None
    assert int(
        Utils.GetLPID(
            'brokendefault',
            workingPrintGroupName,
            False)) > 0
    assert Utils.GetLPID(
        'brokendefault',
        workingPrintGroupName,
        False) is not None

    # test blacklist works
    assert Utils.GetLPID(
        workingPrintGroupName,
        'brokenalternative',
        True,
        [workingPrintGroupName,
         'brokendefault',
         'adm',
         'wheel',
         'root'],
        True) is None

def test_showVersion():
    assert Utils.ShowVersion("12345") == False
    sys.argv = ['testfile', 'version']
    with pytest.raises(SystemExit):
        Utils.ShowVersion("12345")

def test_readFile():
    Utils.WriteFile('/tmp/testfile', 'data')
    assert Utils.ReadFile('/tmp/testfile') == 'data'
    assert Utils.ReadFile('/tmp/filethatdoesntexist') is None
    os.unlink('/tmp/testfile')

def test_writeFile():
    Utils.WriteFile('/tmp/testfile', 'data') == True
    Utils.WriteFile('/tmp/testfile/dsadsaasd', 'data') == False
    os.unlink('/tmp/testfile')

def test_base64encode():
    assert Utils.Base64Encode('data', 'pdf') == 'data:application/pdf;base64,ZGF0YQ=='
    assert Utils.Base64Encode('data', 'other') == 'data:application/octet-stream;base64,ZGF0YQ=='
    assert Utils.Base64Encode('data', 'something') == 'data:application/octet-stream;base64,ZGF0YQ=='

def test_GetLanguage():
    assert Utils.GetLanguage(['en_GB',]) == "en"
    assert Utils.GetLanguage(['en_US',]) == "en"
    assert Utils.GetLanguage(['fr_CA',]) == "fr"
    assert Utils.GetLanguage(['fr_FR',]) == "fr"
    assert Utils.GetLanguage(['it_IT',]) == "it"
    assert Utils.GetLanguage(['en',]) == "en"
    assert Utils.GetLanguage([None,None]) == "en"
    
def test_GetDefaultPaperType():
    assert Utils.GetDefaultPaperType(['en_GB',]) == "A4"
    assert Utils.GetDefaultPaperType(['en_US',]) == "Letter"
    assert Utils.GetDefaultPaperType([None,None]) == "Letter"

def test_GetWindowSize():
    # expect this to fail gracefully if no tty
    assert Utils.GetWindowSize() == None

    # pass in dummy winsize struct
    dummywinsize = struct.pack('HHHH', 800, 600, 0, 0)
    assert Utils.GetWindowSize(dummywinsize) == (800,600)

    # ensure window size of 0x0 returns none
    dummywinsize = struct.pack('HHHH', 0, 0, 0, 0)
    assert Utils.GetWindowSize(dummywinsize) == None

def test_GetTempFileName():
    assert Utils.GetTempFileName() != None
    assert Utils.GetTempFileName().count('/') == 2



#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2014 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

if __name__ == '__main__':  # pragma: no cover

    import os
    import sys
    from auth import Auth
    from ccputils import Utils
    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    # dont refresh tokens if no config
    if os.path.exists(Auth.config):
        requestors, storage = Auth.SetupAuth(False)
    else:
        sys.stderr.write("Missing %s, not attempting to renew tokens\n" % Auth.config)



#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


def printPrinters(printers):
    """Prints display name of printers.

    Formats as multiple columns if possible. Enumerates each printer name
    with integers starting with 1 (not zero).

    Args:
        printers: List of printers.

    Returns:
        number of printers printed
    """

    printer_names = \
        ["%d) %s" % (i + 1, printer['displayName']) for i, printer in enumerate(printers)]

    window_size = Utils.GetWindowSize()
    if window_size is None or window_size[0] > len(printer_names):
        for printer_name in printer_names:
            print printer_name
    else:
        window_width = window_size[1]
        max_name_length = max((len(printer_name) for printer_name in printer_names))
        # How many columns fit in the window, with one space between columns?
        column_quantity = max(1, (window_width + 1) / (max_name_length + 1))
        row_quantity = int(math.ceil(len(printer_names) / float(column_quantity)))

        for row_i in xrange(row_quantity):
            row_printers = []
            for printer_name in printer_names[row_i::row_quantity]:
                row_printers.append(printer_name.ljust(max_name_length))
            print ' '.join(row_printers)
    return len(printers)

if __name__ == '__main__':  # pragma: no cover
    import cups
    import os
    import json
    import sys
    import math
    from auth import Auth
    from printermanager import PrinterManager
    from ccputils import Utils
    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    unattended = False
    answer = ""

    if len(sys.argv) == 2 and sys.argv[1] == 'unattended':
        print "Running unattended, setting up all printers"
        unattended = True

    if os.path.exists(Auth.config):
        try:
            content_file = open(Auth.config, 'r')
            content = content_file.read()
            data = json.loads(content)
        except Exception:
            # remove old config file
            print "Deleting old configuration file: " + Auth.config
            os.remove(Auth.config)

    while True:
        requestors, storage = Auth.SetupAuth(True)
        print "You currently have these accounts configured: "
        for requestor in requestors:
            print requestor.getAccount()
        if unattended:
            break
        answer = raw_input("Add more accounts (Y/N)? ")
        if not answer.lower().startswith("y"):
            break
        else:
            Auth.AddAccount(storage)

    for requestor in requestors:
        addedCount = 0
        connection = cups.Connection()
        cupsprinters = connection.getPrinters()
        prefix = ""
        printer_manager = PrinterManager(requestor)
        printers = printer_manager.getPrinters()
        if printers is None:
            print "Sorry, no printers were found on your Google Cloud Print account."
            continue

        if unattended:
            answer = "y"
        else:
            answer = raw_input("Add all Google Cloud Print printers from %s to CUPS (Y/N)? " %
                               requestor.getAccount())

        if not answer.lower().startswith("y"):
            answer = 1
            print "Not adding printers automatically"

            while int(answer) != 0:
                maxprinterid = printPrinters(printers)
                answer = raw_input("Add printer (1-%d, 0 to cancel)? " % maxprinterid)
                try:
                    answer = int(answer)
                except ValueError:
                    answer = 0
                if answer == 0:
                    continue
                if answer < 1 or answer > maxprinterid:
                    print "\nPrinter %d not found\n" % answer
                    continue

                ccpprinter = printers[answer - 1]
                print "Adding " + printers[answer - 1]['displayName']
                prefixanswer = raw_input("Use a prefix for name of printer (Y/N)? ")
                if prefixanswer.lower().startswith("y"):
                    prefix = raw_input("Prefix ( e.g. GCP- )? ")
                    if prefix == "":
                        print "Not using prefix"

                printername = prefix + ccpprinter.getDisplayName().encode('ascii', 'replace')
                found = False
                for cupsprinter in cupsprinters:
                    if cupsprinters[cupsprinter]['device-uri'] == ccpprinter.getURI():
                        found = True
                if found:
                    print "\nPrinter with %s already exists\n" % printername
                else:
                    printer_manager.addPrinter(printername, ccpprinter, connection)

            continue

        prefixanswer = ""
        if unattended:
            prefixanswer = "Y"
        else:
            prefixanswer = raw_input("Use a prefix for names of created printers (Y/N)? ")

        if prefixanswer.lower().startswith("y"):
            prefix = ""
            if unattended:
                prefix = "GCP-"
            else:
                prefix = raw_input("Prefix ( e.g. GCP- )? ")

            if prefix == "":
                print "Not using prefix"

        for ccpprinter in printers:
            found = False
            for cupsprinter in cupsprinters:
                if cupsprinters[cupsprinter]['device-uri'] == ccpprinter.getURI():
                    found = True

            if found:
                continue

            printername = prefix + ccpprinter.getDisplayName()

            # check if printer name already exists
            foundbyname = False
            for ccpprinter2 in cupsprinters:
                if printer_manager.sanitizePrinterName(cupsprinters[ccpprinter2]['printer-info']) == \
                        printer_manager.sanitizePrinterName(printername):
                    foundbyname = True
            if foundbyname and not unattended:
                answer = raw_input('Printer %s already exists, supply another name (Y/N)? ' %
                                   printer_manager.sanitizePrinterName(printername))
                if answer.startswith("Y") or answer.startswith("y"):
                    printername = raw_input("New printer name? ")
                else:
                    answer = raw_input("Overwrite %s with new printer (Y/N)? " %
                                       printer_manager.sanitizePrinterName(printername))
                    if answer.lower().startswith("n"):
                        printername = ""
            elif foundbyname and unattended:
                print "Not adding printer %s, as already exists" % printername
                printername = ""

            if printername != "":
                printer_manager.addPrinter(printername, ccpprinter, connection)
                cupsprinters = connection.getPrinters()
                addedCount += 1

        if addedCount > 0:
            print "Added %d new printers to CUPS" % addedCount
        else:
            print "No new printers to add"

    printer_uris = []
    printer_manager = PrinterManager(requestors)
    printers = printer_manager.getPrinters()
    for printer in printers:
        printer_uris.append(printer.getURI())

    # check for printers to prune
    prunePrinters = []
    connection = cups.Connection()
    cupsprinters = connection.getPrinters()

    for cupsprinter in cupsprinters:
        if cupsprinters[cupsprinter]['device-uri'].startswith(Utils._PROTOCOL) \
                and cupsprinters[cupsprinter]['device-uri'] not in printer_uris:
            prunePrinters.append(cupsprinter)

    if len(prunePrinters) > 0:
        print "Found %d printers which no longer exist on cloud print:" % len(prunePrinters)
        for printer in prunePrinters:
            print printer
        answer = raw_input("Remove (Y/N)? ")
        if answer.lower().startswith("y"):
            for printer in prunePrinters:
                connection.deletePrinter(printer)
                print "Deleted", printer
        else:
            print "Not removing old printers"



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import pytest
import logging
import sys
sys.path.insert(0, ".")

from cloudprintrequestor import CloudPrintRequestor

global requestor


def teardown_function(function):
    logging.shutdown()
    reload(logging)


def setup_function(function):
    global requestor
    requestor = CloudPrintRequestor()


def test_requestor():
    global requestor
    requestor.setAccount('testdetails')
    assert requestor.getAccount() == 'testdetails'


def test_request():
    global requestor
    assert requestor.doRequest(
        path="test",
        testResponse=json.dumps("randomstring1233")) == "randomstring1233"
    with pytest.raises(ValueError):
        assert requestor.doRequest(path="test", testResponse="")

    assert requestor.doRequest(
        path="test",
        testResponse=json.dumps("randomstring1233"),
        endpointurl=requestor.CLOUDPRINT_URL) == "randomstring1233"
    with pytest.raises(ValueError):
        assert requestor.doRequest(
            path="test",
            testResponse="",
            endpointurl=requestor.CLOUDPRINT_URL)

    # test doing actual requests, not supplying test reponse data
    # we expect them to fail due to missing auth data
    with pytest.raises(ValueError):
        requestor.doRequest(path="printers")
    with pytest.raises(ValueError):
        requestor.doRequest(path="submit", data="test")
    


#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import cups
import hashlib
import json
import locale
import logging
import mimetools
import re
import urllib
import subprocess
import unicodedata
import sys

from ccputils import Utils


class Printer(object):
    _PPD_TEMPLATE_HEAD = """*PPD-Adobe: "4.3"
*%%%%%%%% PPD file for Cloud Print with CUPS.
*FormatVersion: "4.3"
*FileVersion: "1.0"
*LanguageVersion: English
*LanguageEncoding: ISOLatin1
*cupsLanguages: "%(language)s"
*cupsFilter: "application/vnd.cups-postscript 100 -"
*cupsFilter: "application/vnd.cups-pdf 0 -"
*PCFileName: "%(ppdname)s"
*Product: "(Google Cloud Print)"
*Manufacturer: "Google"
*ModelName: "Google Cloud Print"
*ShortNickName: "Google Cloud Print"
*NickName: "Google Cloud Print, 1.0"
*PSVersion: "(3010.000) 550"
*LanguageLevel: "3"
*ColorDevice: True
*DefaultColorSpace: RGB
*FileSystem: False
*Throughput: "1"
*LandscapeOrientation: Minus90
*TTRasterizer: Type42
*%% Driver-defined attributes...
*1284DeviceID: "%(ieee1284)s;"
*OpenUI *PageSize/Media Size: PickOne
*%(language)s.Translation PageSize/Media Size: ""
*OrderDependency: 10 AnySetup *PageSize
*DefaultPageSize: %(defaultpapertype)s.Fullbleed
*PageSize Letter.Fullbleed/US Letter: "<</PageSize[612 792]/ImagingBBox null>>setpagedevice"
*%(language)s.PageSize Letter.Fullbleed/US Letter: ""
*PageSize Legal.Fullbleed/US Legal: "<</PageSize[612 1008]/ImagingBBox null>>setpagedevice"
*%(language)s.PageSize Legal.Fullbleed/US Legal: ""
*PageSize A4.Fullbleed/A4: "<</PageSize[595 842]/ImagingBBox null>>setpagedevice"
*%(language)s.PageSize A4.Fullbleed/A4: ""
*CloseUI: *PageSize
*OpenUI *PageRegion/Page Region: PickOne
*%(language)s.Translation PageRegion/Page Region: ""
*OrderDependency: 10 AnySetup *PageRegion
*DefaultPageRegion: %(defaultpapertype)s.Fullbleed
*PageRegion Letter.Fullbleed/US Letter: "<</PageSize[612 792]/ImagingBBox null>>setpagedevice"
*%(language)s.PageRegion Letter.Fullbleed/US Letter: ""
*PageRegion Legal.Fullbleed/US Legal: "<</PageSize[612 1008]/ImagingBBox null>>setpagedevice"
*%(language)s.PageRegion Legal.Fullbleed/US Legal: ""
*PageRegion A4.Fullbleed/A4: "<</PageSize[595 842]/ImagingBBox null>>setpagedevice"
*%(language)s.PageRegion A4.Fullbleed/A4: ""
*CloseUI: *PageRegion
*DefaultImageableArea: %(defaultpapertype)s.Fullbleed
*ImageableArea Letter.Fullbleed/US Letter: "0 0 612 792"
*ImageableArea Legal.Fullbleed/US Legal: "0 0 612 1008"
*ImageableArea A4.Fullbleed/A4: "0 0 595 842"
*DefaultPaperDimension: %(defaultpapertype)s.Fullbleed
*PaperDimension Letter.Fullbleed/US Letter: "612 792"
*PaperDimension Legal.Fullbleed/US Legal: "612 1008"
*PaperDimension A4.Fullbleed/A4: "595 842"
"""

    _PPD_TEMPLATE_FOOT = """*DefaultFont: Courier
*Font AvantGarde-Book: Standard "(1.05)" Standard ROM
*Font AvantGarde-BookOblique: Standard "(1.05)" Standard ROM
*Font AvantGarde-Demi: Standard "(1.05)" Standard ROM
*Font AvantGarde-DemiOblique: Standard "(1.05)" Standard ROM
*Font Bookman-Demi: Standard "(1.05)" Standard ROM
*Font Bookman-DemiItalic: Standard "(1.05)" Standard ROM
*Font Bookman-Light: Standard "(1.05)" Standard ROM
*Font Bookman-LightItalic: Standard "(1.05)" Standard ROM
*Font Courier: Standard "(1.05)" Standard ROM
*Font Courier-Bold: Standard "(1.05)" Standard ROM
*Font Courier-BoldOblique: Standard "(1.05)" Standard ROM
*Font Courier-Oblique: Standard "(1.05)" Standard ROM
*Font Helvetica: Standard "(1.05)" Standard ROM
*Font Helvetica-Bold: Standard "(1.05)" Standard ROM
*Font Helvetica-BoldOblique: Standard "(1.05)" Standard ROM
*Font Helvetica-Narrow: Standard "(1.05)" Standard ROM
*Font Helvetica-Narrow-Bold: Standard "(1.05)" Standard ROM
*Font Helvetica-Narrow-BoldOblique: Standard "(1.05)" Standard ROM
*Font Helvetica-Narrow-Oblique: Standard "(1.05)" Standard ROM
*Font Helvetica-Oblique: Standard "(1.05)" Standard ROM
*Font NewCenturySchlbk-Bold: Standard "(1.05)" Standard ROM
*Font NewCenturySchlbk-BoldItalic: Standard "(1.05)" Standard ROM
*Font NewCenturySchlbk-Italic: Standard "(1.05)" Standard ROM
*Font NewCenturySchlbk-Roman: Standard "(1.05)" Standard ROM
*Font Palatino-Bold: Standard "(1.05)" Standard ROM
*Font Palatino-BoldItalic: Standard "(1.05)" Standard ROM
*Font Palatino-Italic: Standard "(1.05)" Standard ROM
*Font Palatino-Roman: Standard "(1.05)" Standard ROM
*Font Symbol: Special "(001.005)" Special ROM
*Font Times-Bold: Standard "(1.05)" Standard ROM
*Font Times-BoldItalic: Standard "(1.05)" Standard ROM
*Font Times-Italic: Standard "(1.05)" Standard ROM
*Font Times-Roman: Standard "(1.05)" Standard ROM
*Font ZapfChancery-MediumItalic: Standard "(1.05)" Standard ROM
*Font ZapfDingbats: Special "(001.005)" Special ROM
*% End of cloudprint.ppd, 04169 bytes."""

    _BACKEND_DESCRIPTION = 'network %s "%s" "%s" "%s"'
    _BACKEND_DESCRIPTION_PLUS_LOCATION = 'network %s "%s" "%s @ %s" "%s" "%s"'

    _IEEE_1284 = 'MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL:%s'

    _DRIVER_DESCRIPTION = '"%s" en "Google" "%s (%s)" "%s"'

    _PPD_NAME = 'cupscloudprint:%s:%s.ppd'

    _RESERVED_CAPABILITY_WORDS = set((
        'Duplex', 'Resolution', 'Attribute', 'Choice', 'ColorDevice', 'ColorModel', 'ColorProfile',
        'Copyright', 'CustomMedia', 'Cutter', 'Darkness', 'DriverType', 'FileName', 'Filter',
        'Filter', 'Finishing', 'Font', 'Group', 'HWMargins', 'InputSlot', 'Installable',
        'LocAttribute', 'ManualCopies', 'Manufacturer', 'MaxSize', 'MediaSize', 'MediaType',
        'MinSize', 'ModelName', 'ModelNumber', 'Option', 'PCFileName', 'SimpleColorProfile',
        'Throughput', 'UIConstraints', 'VariablePaperSize', 'Version', 'Color', 'Background',
        'Stamp', 'DestinationColorProfile', 'JCLToPDFInterpreter', 'APAutoSetupTool',
        'APDialogExtension', 'APHelpBook', 'APICADriver', 'APPrinterIconPath',
        'APPrinterLowInkTool', 'APPrinterPreset', 'APPrinterUtilityPath', 'APScannerOnly',
        'APScanAppBundleID'
    ))

    _RESERVED_CAPABILITY_PREFIXES = (
        'cups'
    )

    _FIXED_OPTION_MAPPINGS = {"psk:JobDuplexAllDocumentsContiguously":
                              {'psk:OneSided': "None",
                                  'psk:TwoSidedShortEdge': "DuplexTumble",
                                  'psk:TwoSidedLongEdge': "DuplexNoTumble"},
                              "psk:PageOrientation":
                              {'psk:Landscape': "Landscape",
                               'psk:Portrait': "Portrait"}
                              }

    _FIXED_CAPABILITY_MAPPINGS = {'ns1:Colors': "ColorModel",
                                  'ns1:PrintQualities': "OutputMode",
                                  'ns1:InputBins': "InputSlot",
                                  'psk:JobDuplexAllDocumentsContiguously': "Duplex",
                                  'psk:PageOrientation': "Orientation"}

    _CONVERTCOMMAND = 'convert'

    _IGNORED_CAPABILITIES = [
        'supported_content_type',
        'vendor_capability'
    ]

    def __init__(self, fields, requestor):
        self._fields = fields
        self._requestor = requestor

    def getAccount(self):
        return self._requestor.getAccount()

    def getRequestor(self):
        return self._requestor

    def _getMimeBoundary(self):
        if not hasattr(self, '_mime_boundary'):
            self._mime_boundary = mimetools.choose_boundary()
        return self._mime_boundary

    def __getitem__(self, key):
        if key == 'capabilities' and key not in self._fields:
            self._fields = self._fetchDetails()
            if key not in self._fields:
                # Make sure to only fetch details once.
                self._fields[key] = None

        return self._fields[key]

    def __contains__(self, key):
        return key in self._fields

    def _fetchDetails(self):
        responseobj = self._requestor.printer(self['id'])
        return responseobj['printers'][0]

    def getURI(self):
        """Generates a URI for the Cloud Print Printer.

        Returns:
          URI for the printer
        """
        account = urllib.quote(self.getAccount().encode('ascii', 'replace'))
        printer_id = urllib.quote(self['id'].encode('ascii', 'replace'))
        return "%s%s/%s" % (Utils._PROTOCOL, account, printer_id)

    def getListDescription(self):
        return '%s - %s - %s' % (
            self.getDisplayName().encode('ascii', 'replace'), self.getURI(), self.getAccount())

    def getLocation(self):
        """Gets the location of the printer, or '' if location not available."""

        # Look for hints of a location tag.
        if 'tags' in self:
            for tag in self['tags']:
                if '=' not in tag:
                    continue
                key, value = tag.split('=', 1)
                if 'location' in key:
                    return value

        return ''

    def getCUPSBackendDescription(self):
        display_name = self.getDisplayName()

        location = self.getLocation()
        if location:
            return self._BACKEND_DESCRIPTION_PLUS_LOCATION % (
                self.getURI(), display_name, display_name, location, self.getIEEE1284(), location)
        else:
            return self._BACKEND_DESCRIPTION % (
                self.getURI(), display_name, display_name, self.getIEEE1284())

    def getCUPSDriverDescription(self):
        name = self.getDisplayName().encode('ascii', 'replace')
        return self._DRIVER_DESCRIPTION % (
            self.getPPDName(), name, self.getAccount(), self.getIEEE1284())

    def getIEEE1284(self):
        return self._IEEE_1284 % self.getURI()

    def getDisplayName(self):
        """Gets a name that carbon-based lifeforms can read.

        For example, "HP LaserJet 2000", not "HP_LaserJet-2000".

        Returns:
          A name that makes sense when displayed to a non-technical user
        """

        if 'displayName' in self and self['displayName']:
            return self['displayName']
        else:
            return self['name']

    def getPPDName(self):
        return self._PPD_NAME % (
            urllib.quote(self.getAccount().encode('ascii', 'replace').replace(' ', '-')),
            self['id'].encode('ascii', 'replace').replace(' ', '-'))

    def generatePPD(self):
        """Generates a PPD string for this printer."""
        defaultlocale = locale.getdefaultlocale()
        language = Utils.GetLanguage(defaultlocale)
        defaultpapertype = Utils.GetDefaultPaperType(defaultlocale)
        ppd = self._PPD_TEMPLATE_HEAD % {
            'language': language, 'defaultpapertype': defaultpapertype,
            'ieee1284': self.getIEEE1284(), 'ppdname': self.getPPDName()}
        if self['capabilities'] is not None:
            addedCapabilities = []
            for capabilityname in self['capabilities']['printer']:
                if capabilityname not in self._IGNORED_CAPABILITIES:
                    capability = {}
                    capability['options'] = self['capabilities']['printer'][capabilityname]
                    capability['name'] = capabilityname
                    originCapabilityName = None
                    internalCapabilityName = \
                        self._getInternalName(capability, 'capability', None, addedCapabilities)
                    addedCapabilities.append(internalCapabilityName)
                    if 'displayName' in capability and len(capability['displayName']) > 0:
                        originCapabilityName = self._sanitizeText(capability['displayName'])
                    elif 'psk:DisplayName' in capability and len(capability['psk:DisplayName']) > 0:
                        originCapabilityName = self._sanitizeText(capability['psk:DisplayName'])
                    else:
                        originCapabilityName = self._sanitizeText(capability['name'])

                    newppddata = '*OpenUI *%s/%s: PickOne\n' % \
                        (internalCapabilityName, internalCapabilityName)
                    # translation of capability, allows use of 8
                    # bit chars
                    newppddata += '*%s.Translation %s/%s: ""\n' % \
                        (language, internalCapabilityName, originCapabilityName)
                    addedOptions = []

                    if 'options' in capability and 'option' in capability['options']:
                        for option in capability['options']['option']:
                            originOptionName = None
                            if 'display_name' in option and len(option['display_name']) > 0:
                                originOptionName = self._sanitizeText(option['display_name'])
                            elif 'custom_display_name' in option \
                                    and len(option['custom_display_name']) > 0:
                                originOptionName = self._sanitizeText(option['custom_display_name'])
                            elif 'name' in option and len(option['name']) > 0:
                                originOptionName = self._sanitizeText(option['name'])
                            elif 'type' in option and len(option['type']) > 0:
                                originOptionName = self._sanitizeText(option['type'])
                            else:
                                continue
                            internalOptionName = self._getInternalName(
                                option, 'option', capability['name'], addedOptions)
                            addedOptions.append(internalOptionName)
                            if 'is_default' in option and option['is_default']:
                                newppddata += '*Default%s: %s\n' % (
                                    internalCapabilityName, internalOptionName)
                            newppddata += '*%s %s:%s\n' % \
                                (internalCapabilityName, internalOptionName, internalOptionName)
                            # translation of option, allows use of 8
                            # bit chars
                            value = ''
                            if 'ppd:value' in option:
                                value = option['ppd:value']
                            newppddata += '*%s.%s %s/%s: "%s"\n' % (
                                language,
                                internalCapabilityName,
                                internalOptionName,
                                originOptionName,
                                value)

                        newppddata += '*CloseUI: *%s\n' % internalCapabilityName
                    if len(addedOptions) > 0:
                        ppd += newppddata

        ppd += self._PPD_TEMPLATE_FOOT
        return ppd

    @staticmethod
    def _sanitizeText(text, checkReserved=False):
        sanitisedName = re.sub(r'(:|;| )', '_', text).replace('/', '-').encode('utf8', 'ignore')
        sanitisedName = "".join(ch for ch in unicode(sanitisedName, errors='ignore')
                                if unicodedata.category(ch)[0] != "C")
        if checkReserved and (sanitisedName in Printer._RESERVED_CAPABILITY_WORDS or
                              sanitisedName.startswith(Printer._RESERVED_CAPABILITY_PREFIXES)):
            sanitisedName = 'GCP_' + sanitisedName
        return sanitisedName

    @staticmethod
    def _getInternalName(details, internalType, capabilityName=None, existingList=None):
        returnValue = None
        if existingList is None:
            existingList = []

        fixedNameMap = {}
        # use fixed options for options we recognise
        if internalType == "option":
            # option
            if capabilityName in Printer._FIXED_OPTION_MAPPINGS:
                fixedNameMap = Printer._FIXED_OPTION_MAPPINGS[capabilityName]
        else:
            # capability
            fixedNameMap = Printer._FIXED_CAPABILITY_MAPPINGS

        for itemName in fixedNameMap:
            if details['name'] == itemName:
                returnValue = fixedNameMap[itemName]
                break

        if 'display_name' in details and len(details['display_name']) > 0:
            name = details['display_name']
        elif 'custom_display_name' in details and len(details['custom_display_name']) > 0:
            name = details['custom_display_name']
        elif 'name' in details and len(details['name']) > 0:
            name = details['name']
        elif 'type' in details and len(details['type']) > 0:
            name = details['type']
        else:
            return None

        sanitisedName = Printer._sanitizeText(name, True)

        # only sanitise, no hash
        if returnValue is None and\
                len(sanitisedName) <= 30 and\
                sanitisedName.decode("utf-8", 'ignore').encode("ascii", "ignore") == sanitisedName:
            returnValue = sanitisedName

        if returnValue is None:
            returnValue = hashlib.sha256(sanitisedName).hexdigest()[:7]

        if returnValue not in existingList:
            return returnValue

        origReturnValue = returnValue

        if "GCP_" + origReturnValue not in existingList:
            return "GCP_" + origReturnValue

        # max 100 rotations, prevent infinite loop
        for i in range(1, 100):
            if returnValue in existingList:
                returnValue = "GCP_" + str(i) + "_" + origReturnValue

        # TODO: need to error if limit hit, or run out of chars allowed etc

        return returnValue

    def _encodeMultiPart(self, fields, file_type='application/xml'):
        """Encodes list of parameters for HTTP multipart format.

        Args:
          fields: list of tuples containing name and value of parameters.
          file_type: string if file type different than application/xml.
        Returns:
          A string to be sent as data for the HTTP post request.
        """
        lines = []
        for (key, value) in fields:
            lines.append('--' + self._getMimeBoundary())
            lines.append('Content-Disposition: form-data; name="%s"' % key)
            lines.append('')  # blank line
            lines.append(str(value))
        lines.append('--%s--' % self._getMimeBoundary())
        lines.append('')  # blank line
        return '\r\n'.join(lines)

    @staticmethod
    def _getOverrideCapabilities(overrideoptionsstring):
        overrideoptions = overrideoptionsstring.split(' ')
        overridecapabilities = {}

        ignorecapabilities = ['Orientation']
        for optiontext in overrideoptions:
            if '=' in optiontext:
                optionparts = optiontext.split('=')
                option = optionparts[0]
                if option in ignorecapabilities:
                    continue

                value = optionparts[1]
                overridecapabilities[option] = value

            # landscape
            if optiontext == 'landscape' or optiontext == 'nolandscape':
                overridecapabilities['Orientation'] = 'Landscape'

        return overridecapabilities

    @staticmethod
    def _getCapabilitiesDict(attrs, printercapabilities, overridecapabilities):
        # example:
        # {"version":"1.0","print":{"media_size":{"width_microns":215000,"height_microns":275000,"vendor_id":"15"}}}
        capabilities = {"version": "1.0", "print": {}}
        for attr in attrs:
            if attr['name'].startswith('Default'):
                # gcp setting, reverse back to GCP capability
                gcpname = None
                hashname = attr['name'].replace('Default', '')
                parammap = {}
                valuemap = {}
                deleteparams = ['name', 'custom_name', 'is_default',
                                'display_name', 'custom_display_name']

                # find item name from hashes
                gcpoption = None
                addedCapabilities = []
                for capabilityname in printercapabilities:
                    capability = {}
                    capability['options'] = printercapabilities[capabilityname]
                    capability['name'] = capabilityname
                    if hashname == Printer._getInternalName(capability, 'capability'):
                        gcpname = capability['name']
                        for option in capability['options']['option']:
                            paramname = 'type'
                            if 'type' not in option:
                                if 'vendor_id' in option:
                                    paramname = 'vendor_id'
                            if paramname in option:
                                internalCapability = Printer._getInternalName(
                                    option, 'option', gcpname, addedCapabilities)
                                addedCapabilities.append(internalCapability)
                                if attr['value'] == internalCapability:
                                    gcpoption = option[paramname]
                                    parammap[gcpoption] = paramname

                                    for deleteparam in deleteparams:
                                        if deleteparam in option:
                                            del option[deleteparam]

                                    valuemap[gcpoption] = option
                                    break
                        addedOptions = []
                        for overridecapability in overridecapabilities:
                            if 'Default' + overridecapability == attr['name']:
                                selectedoption = overridecapabilities[
                                    overridecapability]
                                for option in capability['options']['option']:
                                    paramname = 'type'
                                    if 'type' not in option:
                                        if 'vendor_id' in option:
                                            paramname = 'vendor_id'
                                    if paramname in option:
                                        internalOption = Printer._getInternalName(
                                            option, 'option', gcpname, addedOptions)
                                        addedOptions.append(internalOption)
                                        if selectedoption == internalOption:
                                            gcpoption = option[paramname]
                                            parammap[gcpoption] = paramname

                                            for deleteparam in deleteparams:
                                                if deleteparam in option:
                                                    del option[deleteparam]

                                            valuemap[gcpoption] = option
                                            break
                                break
                        break

                if gcpname is not None and gcpoption is not None:
                    if parammap[gcpoption] == 'vendor_id':
                        capabilities['print'][gcpname] = valuemap[gcpoption]
                    else:
                        capabilities['print'][gcpname] = {parammap[gcpoption]: gcpoption}
        return capabilities

    @staticmethod
    def _attrListToArray(attrs):
        return ({'name': attr.name, 'value': attr.value} for attr in attrs)

    def _getCapabilities(self, cupsprintername, overrideoptionsstring):
        """Gets capabilities of printer and maps them against list

        Args:
          overrideoptionsstring: override for print job
        Returns:
          List of capabilities
        """
        connection = cups.Connection()
        overridecapabilities = self._getOverrideCapabilities(overrideoptionsstring)
        overrideDefaultDefaults = {'Duplex': 'None'}

        for capability in overrideDefaultDefaults:
            if capability not in overridecapabilities:
                overridecapabilities[capability] = overrideDefaultDefaults[capability]
        attrs = cups.PPD(connection.getPPD(cupsprintername)).attributes
        attrArray = self._attrListToArray(attrs)
        return self._getCapabilitiesDict(attrArray,
                                         self['capabilities']['printer'],
                                         overridecapabilities)

    def submitJob(self, jobtype, jobfile, jobdata, jobname, cupsprintername, options=""):
        """Submits a job to printerid with content of dataUrl.

        Args:
          jobtype: string, must match the dictionary keys in content and content_type.
          jobfile: string, points to source for job. Could be a pathname or id string.
          jobdata: string, data for print job
          jobname: string, name of the print job ( usually page name ).
          options: string, key-value pair of options from print job.

        Returns:
          True if submitted, False otherwise
        """
        rotate = 0

        # refuse to submit empty jobdata
        if len(jobdata) == 0:
            sys.stderr.write("ERROR: Job data is empty\n")
            return False

        if jobfile is None or jobfile == "":
            jobfile = "Unknown"

        for optiontext in options.split(' '):

            # landscape
            if optiontext == 'landscape':
                # landscape
                rotate = 90

            # nolandscape - already rotates
            if optiontext == 'nolandscape':
                # rotate back
                rotate = 270

        if jobtype not in ['png', 'jpeg', 'pdf']:
            sys.stderr.write("ERROR: Unknown job type: %s\n" % jobtype)
            return False
        else:
            if rotate != 0:
                try:
                    command = [self._CONVERTCOMMAND, '-density', '300x300', '-',
                               '-rotate', str(rotate), '-']
                    p = subprocess.Popen(command, stdout=subprocess.PIPE,
                                         stdin=subprocess.PIPE, stderr=subprocess.PIPE)
                    newjobdata = p.communicate(jobdata)
                    if p.returncode == 0:
                        jobdata = newjobdata[0]
                    else:
                        logging.error("Failed to rotate: %s" % newjobdata[1])
                        return False
                except Exception as error_msg:
                    logging.error("Convert command errored when rotating: %s" % str(error_msg))
                    return False

        if jobname == "":
            title = "Untitled page"
        else:
            title = jobname
        headers = [
            ('printerid', self['id']),
            ('title', title),
            ('content', Utils.Base64Encode(jobdata, jobtype)),
            ('contentType', 'dataUrl'),
            ('ticket', json.dumps(self._getCapabilities(cupsprintername, options)))
        ]
        logging.info('Capability headers are: %s', headers[4])
        data = self._encodeMultiPart(headers, 'dataUrl')

        try:
            responseobj = self.getRequestor().submit(data, self._getMimeBoundary())
            if responseobj['success']:
                return True
            else:
                sys.stderr.write("ERROR: Error response from Cloud Print for type %s: %s\n" %
                                 (jobtype, responseobj['message']))
                return False

        except Exception as error_msg:
            sys.stderr.write("ERROR: Print job %s failed with %s\n" % (jobtype, error_msg))
            return False



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for Google Compute Engine

Utilities for making it easier to use OAuth 2.0 on Google Compute Engine.
"""

import json
import logging
from six.moves import urllib

from oauth2client._helpers import _from_bytes
from oauth2client import util
from oauth2client.client import HttpAccessTokenRefreshError
from oauth2client.client import AssertionCredentials


__author__ = 'jcgregorio@google.com (Joe Gregorio)'

logger = logging.getLogger(__name__)

# URI Template for the endpoint that returns access_tokens.
META = ('http://metadata.google.internal/0.1/meta-data/service-accounts/'
        'default/acquire{?scope}')


class AppAssertionCredentials(AssertionCredentials):
    """Credentials object for Compute Engine Assertion Grants

    This object will allow a Compute Engine instance to identify itself to
    Google and other OAuth 2.0 servers that can verify assertions. It can be
    used for the purpose of accessing data stored under an account assigned to
    the Compute Engine instance itself.

    This credential does not require a flow to instantiate because it
    represents a two legged flow, and therefore has all of the required
    information to generate and refresh its own access tokens.
    """

    @util.positional(2)
    def __init__(self, scope, **kwargs):
        """Constructor for AppAssertionCredentials

        Args:
            scope: string or iterable of strings, scope(s) of the credentials
                   being requested.
        """
        self.scope = util.scopes_to_string(scope)
        self.kwargs = kwargs

        # Assertion type is no longer used, but still in the
        # parent class signature.
        super(AppAssertionCredentials, self).__init__(None)

    @classmethod
    def from_json(cls, json_data):
        data = json.loads(_from_bytes(json_data))
        return AppAssertionCredentials(data['scope'])

    def _refresh(self, http_request):
        """Refreshes the access_token.

        Skip all the storage hoops and just refresh using the API.

        Args:
            http_request: callable, a callable that matches the method
                          signature of httplib2.Http.request, used to make
                          the refresh request.

        Raises:
            HttpAccessTokenRefreshError: When the refresh fails.
        """
        query = '?scope=%s' % urllib.parse.quote(self.scope, '')
        uri = META.replace('{?scope}', query)
        response, content = http_request(uri)
        content = _from_bytes(content)
        if response.status == 200:
            try:
                d = json.loads(content)
            except Exception as e:
                raise HttpAccessTokenRefreshError(str(e),
                                                  status=response.status)
            self.access_token = d['accessToken']
        else:
            if response.status == 404:
                content += (' This can occur if a VM was created'
                            ' with no service account or scopes.')
            raise HttpAccessTokenRefreshError(content, status=response.status)

    @property
    def serialization_data(self):
        raise NotImplementedError(
            'Cannot serialize credentials for GCE service accounts.')

    def create_scoped_required(self):
        return not self.scope

    def create_scoped(self, scopes):
        return AppAssertionCredentials(scopes, **self.kwargs)



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import urllib
import cups
import os
import stat
import grp
import pytest
import logging
import sys
sys.path.insert(0, ".")

from auth import Auth
from mockrequestor import MockRequestor
from oauth2client import client
from oauth2client import multistore_file
from ccputils import Utils


def setup_function(function):
    # setup mock requestors
    global requestors
    requestors = []

    # account without special chars
    mockRequestorInstance1 = MockRequestor()
    mockRequestorInstance1.setAccount('testaccount1')
    mockRequestorInstance1.printers = []
    requestors.append(mockRequestorInstance1)

    Auth.config = '/tmp/cloudprint.conf'


def teardown_function(function):
    if os.path.exists(Auth.config):
        os.unlink(Auth.config)
    logging.shutdown()
    reload(logging)


def test_fixConfigPermissions():
    configfile = open(Auth.config, "w")
    configfile.close()

    os.chmod(Auth.config, 0000)
    assert '0000' == oct(os.stat(Auth.config)[stat.ST_MODE])[-4:]
    assert True == Utils.FixFilePermissions(Auth.config)[0]
    assert '0660' == oct(os.stat(Auth.config)[stat.ST_MODE])[-4:]

    origconfig = Auth.config
    Auth.config = '/tmp/filethatdoesntexist'
    assert (False, False) == Utils.FixFilePermissions(Auth.config)
    Auth.config = origconfig


@pytest.mark.skipif(
    grp.getgrnam('lp').gr_gid not in (os.getgroups()) and os.getuid() != 0,
    reason="will only pass if running user part of lp group or root")
def test_fixConfigOwnerships():
    configfile = open(Auth.config, "w")
    configfile.close()

    assert Utils.GetLPID() != os.stat(Auth.config).st_gid
    assert True == Utils.FixFilePermissions(Auth.config)[1]
    assert Utils.GetLPID() == os.stat(Auth.config).st_gid


def test_setupAuth():
    testUserName = 'testaccount1'

    # create initial file
    assert os.path.exists(Auth.config) == False
    assert Auth.SetupAuth(False) == (False, False)
    assert os.path.exists(Auth.config) == True

    # ensure permissions are correct after creating config
    assert '0660' == oct(os.stat(Auth.config)[stat.ST_MODE])[-4:]

    # add dummy details
    storage = multistore_file.get_credential_storage(
        Auth.config,
        Auth.clientid,
        'testuseraccount',
        ['https://www.googleapis.com/auth/cloudprint'])

    credentials = client.OAuth2Credentials('test', Auth.clientid,
                                           'testsecret', 'testtoken', 1,
                                           'https://www.googleapis.com/auth/cloudprint', testUserName)
    storage.put(credentials)

    # ensure permissions are correct after populating config
    assert '0660' == oct(os.stat(Auth.config)[stat.ST_MODE])[-4:]

    # re-run to test getting credentials
    requestors, storage = Auth.SetupAuth(False)
    assert requestors is not None
    assert storage is not None

    # check deleting account
    assert Auth.DeleteAccount(testUserName) is None
    requestors, storage = Auth.SetupAuth(False)
    assert requestors == False
    assert storage == False

def test_setupAuthInteractive():
    
    # ensure running setup in interactive mode tries to read stdin
    with pytest.raises(IOError):
        Auth.SetupAuth(True)

def test_renewToken():
    global requestors
    storage = multistore_file.get_credential_storage(
        Auth.config,
        Auth.clientid,
        'testuseraccount',
        ['https://www.googleapis.com/auth/cloudprint'])
    
    credentials = client.OAuth2Credentials('test', Auth.clientid,
                                           'testsecret', 'testtoken', 1,
                                           'https://www.googleapis.com/auth/cloudprint', 'testaccount1')
    
    # test renewing token exits in non-interactive mode
    with pytest.raises(SystemExit):
        Auth.RenewToken(False, requestors[0], credentials,storage, 'test')
        
    # test renewing interactively tries to read from stdin
    with pytest.raises(IOError):
        assert Auth.RenewToken(True, requestors[0], credentials,storage, 'test') == False

@pytest.mark.skipif(
    grp.getgrnam('lp').gr_gid not in (os.getgroups()) and os.getuid() != 0,
    reason="will only pass if running user part of lp group or root")
def test_setupAuthOwnership():
    assert Auth.SetupAuth(False) == (False, False)

    # ensure ownership is correct after creating config
    assert Utils.GetLPID() == os.stat(Auth.config).st_gid

    # add dummy details
    storage = multistore_file.get_credential_storage(
        Auth.config,
        Auth.clientid,
        'testuseraccount',
        ['https://www.googleapis.com/auth/cloudprint'])

    credentials = client.OAuth2Credentials('test', Auth.clientid,
                                           'testsecret', 'testtoken', 1,
                                           'https://www.googleapis.com/auth/cloudprint', 'testaccount1')
    storage.put(credentials)

    # ensure ownership is correct after populating config
    assert Utils.GetLPID() == os.stat(Auth.config).st_gid



#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

def getDriveFiles(requestors):
    returnValue = []
    for requestor in requestors:
        responseobj = requestor.doRequest(
            'files', endpointurl="https://www.googleapis.com/drive/v2")
        if 'error' in responseobj:
            print "Errored fetching files from drive"
        else:
            for item in responseobj['items']:
                returnValue.append(item)
    if len(returnValue) == 0:
        return None
    return returnValue

if __name__ == '__main__':  # pragma: no cover
    import sys
    sys.path.insert(0, ".")

    from auth import Auth
    from ccputils import Utils
    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    requestors, storage = Auth.SetupAuth(True,
        permissions=['https://www.googleapis.com/auth/cloudprint', 'https://www.googleapis.com/auth/drive.readonly'])
    files = getDriveFiles(requestors)
    if files is None:
        print "No Files Found"
        sys.exit(1)

    for drivefile in files:
        if len(sys.argv) == 2 and drivefile['title'] == sys.argv[1] + '.pdf':
            print drivefile['fileSize']
            sys.exit(0)
        elif len(sys.argv) != 2:
            print drivefile['title']



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import httplib2
import json
import logging
import time


class CloudPrintRequestor(httplib2.Http):

    CLOUDPRINT_URL = 'https://www.google.com/cloudprint'
    account = None

    def setAccount(self, account):
        """Sets the account name

        Args:
          filename: string, name of the account
        """
        self.account = account

    def getAccount(self):
        """Gets the account name

        Return:
          string: Account name.
        """
        return self.account

    def doRequest(self, path, headers=None, data=None, boundary=None, testResponse=None,
                  endpointurl=None):
        """Sends a request to Google Cloud Print

        Args:
          path: string, path part of url
          headers: list, headers to send to GCP
          data: string, body part of request
          boundary: string, boundary part of http forms
        Return:
          list: Decoded json response from Google.
        """

        rpc_name = path.split('?', 1)[0]
        logging.debug('Calling %s RPC' % rpc_name)
        start_time = time.time()

        # force useragent to CCP
        if headers is None:
            headers = {}
        headers['user-agent'] = "CUPS Cloud Print"

        url = '%s/%s' % (self.CLOUDPRINT_URL, path)
        if endpointurl is not None:
            url = '%s/%s' % (endpointurl, path)

        # use test response for testing
        if testResponse is None:
            if data is None:
                headers, response = self.request(url, method="GET", headers=headers)
            else:
                headers['Content-Length'] = str(len(data))
                contenttype = 'multipart/form-data;boundary=%s' % boundary
                headers['Content-Type'] = contenttype
                headers, response = self.request(
                    url, method="POST", body=data, headers=headers)
        else:
            response = testResponse

        try:
            decodedresponse = json.loads(response)
        except ValueError as e:
            print "ERROR: Failed to decode JSON, value was: " + response
            raise e

        logging.debug('%s RPC took %.2f seconds' % (rpc_name, time.time() - start_time))

        return decodedresponse

    def search(self):
        return self.doRequest('search?connection_status=ALL&client=webui')

    def printer(self, printerid):
        return self.doRequest('printer?printerid=%s' % printerid)

    def submit(self, edata, boundary):
        return self.doRequest('submit', data=edata, boundary=boundary)



#! /bin/sh
"true" '''\'
if command -v python2 > /dev/null; then
  exec python2 "$0" "$@"
else
  exec python "$0" "$@"
fi
exit $?
'''

#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2013 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

if __name__ == '__main__':  # pragma: no cover

    import sys
    import os
    import subprocess

    libpaths = [
        "/Library/cloudprint-cups/",
        "/usr/local/share/cloudprint-cups/",
        "/usr/share/cloudprint-cups"
    ]
    addedPath = False
    for libpath in libpaths:
        if os.path.exists(libpath):
            sys.path.insert(0, libpath)
            addedPath = True
            break

    if not addedPath:
        sys.stderr.write("ERROR: Could not find any valid path for python files\n")
        sys.exit(1)

    from auth import Auth
    from printermanager import PrinterManager
    from ccputils import Utils
    Utils.SetupLogging()

    # line below is replaced on commit
    CCPVersion = "20140814.2 000000"
    Utils.ShowVersion(CCPVersion)

    requestors, storage = Auth.SetupAuth(True)
    printer_manager = PrinterManager(requestors)
    printers = printer_manager.getPrinters()
    if printers is None:
        print "ERROR: No Printers Found"
        sys.exit(1)

    for printer in printers:
        print printer.getCUPSDriverDescription()
        print ""
        print printer._fields
        print printer['capabilities']
        print "\n"
        ppdname = printer.getPPDName()
        p1 = subprocess.Popen(
            (os.path.join(libpath, 'dynamicppd.py'), 'cat', ppdname.lstrip('-')),
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE)
        ppddata = p1.communicate()[0]
        p = subprocess.Popen(['cupstestppd', '-'], stdout=subprocess.PIPE,
                             stdin=subprocess.PIPE, stderr=subprocess.PIPE)
        testdata = p.communicate(ppddata)[0]
        result = p.returncode
        print "Result of cupstestppd was " + str(result)
        print "".join(testdata)
        if result != 0:
            print "cupstestppd errored: "
            print ppddata
            print "\n"



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import cups
import urllib
import mimetools
import re
from auth import Auth
from urlparse import urlparse
from printer import Printer
from ccputils import Utils


class PrinterManager:
    BOUNDARY = mimetools.choose_boundary()
    CRLF = '\r\n'
    requestors = None
    requestor = None
    cachedPrinterDetails = {}
    reservedCapabilityWords = set((
        'Duplex', 'Resolution', 'Attribute', 'Choice', 'ColorDevice', 'ColorModel', 'ColorProfile',
        'Copyright', 'CustomMedia', 'Cutter', 'Darkness', 'DriverType', 'FileName', 'Filter',
        'Filter', 'Finishing', 'Font', 'Group', 'HWMargins', 'InputSlot', 'Installable',
        'LocAttribute', 'ManualCopies', 'Manufacturer', 'MaxSize', 'MediaSize', 'MediaType',
        'MinSize', 'ModelName', 'ModelNumber', 'Option', 'PCFileName', 'SimpleColorProfile',
        'Throughput', 'UIConstraints', 'VariablePaperSize', 'Version', 'Color', 'Background',
        'Stamp', 'DestinationColorProfile'
    ))
    URIFormatLatest = 9999
    URIFormat20140308 = 3
    URIFormat20140307 = 2
    URIFormat20140210 = 1
    backendDescription =\
        'network %s "%s" "Google Cloud Print" "MFG:Google;MDL:Cloud Print;DES:GoogleCloudPrint;"'

    def __init__(self, requestors):
        """Create an instance of PrinterManager, with authorised requestor

        Args:
          requestors: list or CloudPrintRequestor instance, A list of
          requestors, or a single requestor to use for all Cloud Print
          requests.
        """
        if requestors is not None:
            if isinstance(requestors, list):
                self.requestors = requestors
            else:
                self.requestors = [requestors]

    def getCUPSPrintersForAccount(self, account):
        connection = cups.Connection()
        cupsprinters = connection.getPrinters()
        accountPrinters = []
        for cupsprinter in cupsprinters:
            printer = self.getPrinterByURI(cupsprinters[cupsprinter]['device-uri'])
            if printer is not None:
                if printer.getAccount() == account:
                    accountPrinters.append(cupsprinters[cupsprinter])
        return accountPrinters, connection

    def getPrinter(self, printerId, accountName):
        """Fetch one printer, including capabilities.

        Args:
          printerId: something like e64b1063-80e7-a87e-496c-3caa8cb7d736
          accountName: email address (account) printer is associated with

        Returns:
          A Printer object, or None if printer not found."""

        for requestor in self.requestors:
            if accountName != requestor.getAccount():
                continue

            response = requestor.printer(printerId)
            if response is None or not response['success'] \
                    or 'printers' not in response or not response['printers']:
                break

            return Printer(response['printers'][0], requestor)

        return None

    def getPrinters(self, accountName=None):
        """Fetch a list of printers

        Returns:
          list: list of printers for the accounts.
        """
        if not hasattr(self, '_printers'):
            self._printers = []
            for requestor in self.requestors:
                if accountName is not None and accountName != requestor.getAccount():
                    continue

                responseobj = requestor.search()
                if 'printers' in responseobj:
                    for printer_info in responseobj['printers']:
                        self._printers.append(Printer(printer_info, requestor))

        return self._printers

    def sanitizePrinterName(self, name):
        """Sanitizes printer name for CUPS

        Args:
          name: string, name of printer from Google Cloud Print

        Returns:
          string: CUPS-friendly name for the printer
        """
        return re.sub('[^a-zA-Z0-9\-_]', '', name.encode('ascii', 'replace').replace(' ', '_'))

    def addPrinter(self, printername, printer, connection, ppd=None):
        """Adds a printer to CUPS

        Args:
          printername: string, name of the printer to add
          uri: string, uri of the Cloud Print device
          connection: connection, CUPS connection

        Returns:
          None
        """
        # fix printer name
        printername = self.sanitizePrinterName(printername)
        result = None
        printerppdname = None
        try:
            if ppd is None:
                printerppdname = printer.getPPDName()
            else:
                printerppdname = ppd
            location = printer.getLocation()
            if not location:
                location = 'Google Cloud Print'

            result = connection.addPrinter(
                name=printername, ppdname=printerppdname, info=printername,
                location=location, device=printer.getURI())
            connection.enablePrinter(printername)
            connection.acceptJobs(printername)
            connection.setPrinterShared(printername, False)
        except Exception as error:
            result = error
        if result is None:
            print "Added " + printername
            return True
        else:
            print "Error adding: " + printername, result
            return False

    @staticmethod
    def _getAccountNameAndPrinterIdFromURI(uri):
        splituri = uri.rsplit('/', 2)
        accountName = urllib.unquote(splituri[1])
        printerId = urllib.unquote(splituri[2])
        return accountName, printerId

    def parseLegacyURI(self, uristring, requestors):
        """Parses previous CUPS Cloud Print URIs, only used for upgrades

        Args:
          uristring: string, uri of the Cloud Print device

        Returns:
          string: account name
          string: google cloud print printer name
          string: google cloud print printer id
          int: format id
        """
        formatId = None
        printerName = None
        accountName = None
        printerId = None
        uri = urlparse(uristring)
        pathparts = uri.path.strip('/').split('/')
        if uri.scheme == Utils._OLD_PROTOCOL_NAME:
            if len(pathparts) == 2:
                formatId = PrinterManager.URIFormat20140307
                printerId = urllib.unquote(pathparts[1])
                accountName = urllib.unquote(pathparts[0])
                printerName = urllib.unquote(uri.netloc)
            else:
                if urllib.unquote(uri.netloc) not in Auth.GetAccountNames(requestors):
                    formatId = PrinterManager.URIFormat20140210
                    printerName = urllib.unquote(uri.netloc)
                    accountName = urllib.unquote(pathparts[0])
                else:
                    formatId = PrinterManager.URIFormat20140308
                    printerId = urllib.unquote(pathparts[0])
                    printerName = None
                    accountName = urllib.unquote(uri.netloc)
        elif uri.scheme == Utils._PROTOCOL_NAME:
            formatId = PrinterManager.URIFormatLatest
            printerId = urllib.unquote(pathparts[0])
            printerName = None
            accountName = urllib.unquote(uri.netloc)

        return accountName, printerName, printerId, formatId

    def findRequestorForAccount(self, account):
        """Searches the requestors in the printer object for the requestor for a specific account

        Args:
          account: string, account name
        Return:
          requestor: Single requestor object for the account, or None if no account found
        """
        for requestor in self.requestors:
            if requestor.getAccount() == account:
                return requestor

    def getPrinterByURI(self, uri):
        accountName, printerId = self._getAccountNameAndPrinterIdFromURI(uri)
        return self.getPrinter(printerId, accountName)

    def getPrinterIDByDetails(self, account, printername, printerid):
        """Gets printer id and requestor by printer

        Args:
          uri: string, printer uri
        Return:
          printer id: Single requestor object for the account, or None if no account found
          requestor: Single requestor object for the account
        """
        # find requestor based on account
        requestor = self.findRequestorForAccount(urllib.unquote(account))
        if requestor is None:
            return None, None

        if printerid is not None:
            return printerid, requestor
        else:
            return None, None



#    CUPS Cloudprint - Print via Google Cloud Print
#    Copyright (C) 2011 Simon Cadman
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import cups
import urllib
import logging
import sys
sys.path.insert(0, ".")

from printer import Printer
from printermanager import PrinterManager
from mockrequestor import MockRequestor
from ccputils import Utils

global requestors, printerManagerInstance


def setup_function(function):
    # setup mock requestors
    global requestors
    requestors = []

    # account without special chars
    mockRequestorInstance1 = MockRequestor()
    mockRequestorInstance1.setAccount('testaccount1')
    mockRequestorInstance1.printers = []
    requestors.append(mockRequestorInstance1)

    # with @ symbol
    mockRequestorInstance2 = MockRequestor()
    mockRequestorInstance2.setAccount('testaccount2@gmail.com')
    mockRequestorInstance2.printers = [{'name': 'Save to Google Drive',
                                        'id': '__test_save_docs',
                                        'capabilities': [{'name': 'ns1:Colors',
                                                          'type': 'Feature'}]},
                                       ]
    requestors.append(mockRequestorInstance2)

    # 1 letter
    mockRequestorInstance3 = MockRequestor()
    mockRequestorInstance3.setAccount('t')
    mockRequestorInstance3.printers = []
    requestors.append(mockRequestorInstance3)

    # instantiate printer item
    if function != test_instantiate:
        test_instantiate()


def teardown_function(function):
    global requestors
    requestors = None
    logging.shutdown()
    reload(logging)

def test_parseURI():
    global printerManagerInstance, requestors
    accountName, printerid = printerManagerInstance._getAccountNameAndPrinterIdFromURI(
        Utils._PROTOCOL + "testaccount2%40gmail.com/testid")
    assert printerid == "testid"
    assert accountName == "testaccount2@gmail.com"

def test_parseLegacyURI():
    global printerManagerInstance, requestors

    # 20140210 format
    account, printername, printerid, formatid = printerManagerInstance.parseLegacyURI(
        Utils._OLD_PROTOCOL + "printername/testaccount2%40gmail.com/", requestors)
    assert formatid == printerManagerInstance.URIFormat20140210
    assert account == "testaccount2@gmail.com"
    assert printername == "printername"
    assert printerid is None

    # 20140307 format
    account, printername, printerid, formatid = printerManagerInstance.parseLegacyURI(
        Utils._OLD_PROTOCOL + "printername/testaccount2%40gmail.com/testid", requestors)
    assert formatid == printerManagerInstance.URIFormat20140307
    assert account == "testaccount2@gmail.com"
    assert printername == "printername"
    assert printerid == "testid"

    # 20140308 format
    account, printername, printerid, formatid = printerManagerInstance.parseLegacyURI(
        Utils._OLD_PROTOCOL + "testaccount2%40gmail.com/testid", requestors)
    assert formatid == printerManagerInstance.URIFormat20140308
    assert account == "testaccount2@gmail.com"
    assert printerid == "testid"
    assert printername is None
    
    # 20140621+ format
    account, printername, printerid, formatid = printerManagerInstance.parseLegacyURI(
        Utils._PROTOCOL + "testaccount2%40gmail.com/testid", requestors)
    assert formatid == printerManagerInstance.URIFormatLatest
    assert account == "testaccount2@gmail.com"
    assert printerid == "testid"
    assert printername is None

def test_getPrinterIDByDetails():
    printerid, requestor = printerManagerInstance.getPrinterIDByDetails(
        "testaccount2@gmail.com", "printername", "testid")
    assert printerid == "testid"
    assert isinstance(requestor, MockRequestor)
    assert requestor.getAccount() == 'testaccount2@gmail.com'
    
    # test fails
    printerid, requestor = printerManagerInstance.getPrinterIDByDetails(
        "accountthatdoesntexist", "printernamethatdoesntexist", "testidthatdoesntexist")
    assert printerid is None
    assert requestor is None
    
    printerid, requestor = printerManagerInstance.getPrinterIDByDetails(
        "testaccount2@gmail.com", "printernamethatdoesntexist", None)
    assert printerid is None
    assert requestor is None


def test_getCUPSPrintersForAccount():
    global printerManagerInstance, requestors

    foundprinters, connection = printerManagerInstance.getCUPSPrintersForAccount(
        requestors[0].getAccount())
    assert foundprinters == []
    assert isinstance(connection, cups.Connection)

    # total printer
    totalPrinters = 0
    for requestor in requestors:
        totalPrinters += len(requestor.printers)

    fullprintersforaccount = printerManagerInstance.getPrinters(requestors[1].getAccount())
    assert len(fullprintersforaccount) == len(requestors[1].printers)

    fullprinters = printerManagerInstance.getPrinters()
    assert len(fullprinters) == totalPrinters

    printers = printerManagerInstance.getPrinters()
    assert len(printers) == totalPrinters
    printer = printers[0]

    # get ppd
    ppdid = 'MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL'
    ppds = connection.getPPDs(ppd_device_id=ppdid)
    printerppdname, printerppd = ppds.popitem()

    # test add printer to cups
    assert printerManagerInstance.addPrinter(
        printer['name'],
        printer,
        connection,
        printerppdname) is not None
    foundprinters, newconnection = printerManagerInstance.getCUPSPrintersForAccount(
        requestors[1].getAccount())
    # delete test printer
    connection.deletePrinter(printerManagerInstance.sanitizePrinterName(printer['name']))

    assert isinstance(foundprinters, list)
    assert len(foundprinters) == 1
    assert isinstance(connection, cups.Connection)

def test_getPrinter():
    global requestors, printerManagerInstance
    assert printerManagerInstance.getPrinter('test', 'missingaccount') == None
    assert isinstance(printerManagerInstance.getPrinter('__test_save_docs', requestors[1].getAccount()),Printer)
    assert printerManagerInstance.getPrinter('test', requestors[0].getAccount()) == None

def test_instantiate():
    global requestors, printerManagerInstance
    # verify adding single requestor works
    printerManagerInstance = PrinterManager(requestors[0])
    assert printerManagerInstance.requestors[0] == requestors[0]
    assert len(printerManagerInstance.requestors) == 1

    # verify adding whole array of requestors works
    printerManagerInstance = PrinterManager(requestors)
    assert printerManagerInstance.requestors == requestors
    assert len(printerManagerInstance.requestors) == len(requestors)

def test_GetPrinterByURIFails():
    global printerManagerInstance, requestors

    # ensure invalid account returns None/None
    printerIdNoneTest = printerManagerInstance.getPrinterByURI(
        Utils._PROTOCOL + 'testprinter/accountthatdoesntexist')
    assert printerIdNoneTest is None

    # ensure invalid printer on valid account returns None/None
    printerIdNoneTest = printerManagerInstance.getPrinterByURI(
        Utils._PROTOCOL + 'testprinter/' + urllib.quote(requestors[0].getAccount()))
    assert printerIdNoneTest is None


def test_addPrinterFails():
    global printerManagerInstance
    assert printerManagerInstance.addPrinter('', None, '') == False

def test_invalidRequest():
    testMock = MockRequestor()
    assert testMock.doRequest('thisrequestisinvalid') is None

def test_printers():
    global printerManagerInstance, requestors

    # test cups connection
    connection = cups.Connection()
    cupsprinters = connection.getPrinters()

    # total printer
    totalPrinters = 0
    for requestor in requestors:
        totalPrinters += len(requestor.printers)
    
    # test getting printers for specific account
    printersforaccount = printerManagerInstance.getPrinters(requestors[1].getAccount())
    assert len(printersforaccount) == len(requestors[1].printers)
    
    printers = printerManagerInstance.getPrinters()
    import re
    assert len(printers) == totalPrinters
    for printer in printers:

        # name
        assert isinstance(printer['name'],basestring)
        assert len(printer['name']) > 0

        # account
        assert isinstance(printer.getAccount(),basestring)
        assert len(printer.getAccount()) > 0

        # id
        assert isinstance(printer['id'],basestring)
        assert len(printer['id']) > 0

        # test encoding and decoding printer details to/from uri
        uritest = re.compile(
            Utils._PROTOCOL + "(.*)/" + urllib.quote(printer['id']))
        assert isinstance(printer.getURI(),basestring)
        assert len(printer.getURI()) > 0
        assert uritest.match(printer.getURI()) is not None

        accountName, printerId = printerManagerInstance._getAccountNameAndPrinterIdFromURI(printer.getURI())
        assert isinstance(accountName,basestring)
        assert isinstance(printerId,basestring)
        assert accountName == printer.getAccount()
        assert printerId == printer['id']

        # get ppd
        ppdid = 'MFG:Google;DRV:GCP;CMD:POSTSCRIPT;DES:GoogleCloudPrint;MDL'
        ppds = connection.getPPDs(ppd_device_id=ppdid)
        printerppdname, printerppd = ppds.popitem()

        # test add printer to cups
        assert printerManagerInstance.addPrinter(
            printer['name'],
            printer,
            connection,
            printerppdname) is not None
        testprintername = printerManagerInstance.sanitizePrinterName(printer['name'])

        # test printer actually added to cups
        cupsPrinters = connection.getPrinters()
        found = False
        for cupsPrinter in cupsPrinters:
            if (cupsPrinters[cupsPrinter]['printer-info'] == testprintername):
                found = True
                break

        assert found == True
        
        # delete test printer
        connection.deletePrinter(testprintername)
